# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from opentelemetry.proto.collector.logs.v1 import logs_service_pb2 as opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2


class LogsServiceStub(object):
    """Service that can be used to push logs between one Application instrumented with
    OpenTelemetry and an collector, or between an collector and a central collector (in this
    case logs are sent/received to/from multiple Applications).
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Export = channel.unary_unary(
                '/opentelemetry.proto.collector.logs.v1.LogsService/Export',
                request_serializer=opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceRequest.SerializeToString,
                response_deserializer=opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceResponse.FromString,
                )


class LogsServiceServicer(object):
    """Service that can be used to push logs between one Application instrumented with
    OpenTelemetry and an collector, or between an collector and a central collector (in this
    case logs are sent/received to/from multiple Applications).
    """

    def Export(self, request, context):
        """For performance reasons, it is recommended to keep this RPC
        alive for the entire life of the application.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_LogsServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Export': grpc.unary_unary_rpc_method_handler(
                    servicer.Export,
                    request_deserializer=opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceRequest.FromString,
                    response_serializer=opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'opentelemetry.proto.collector.logs.v1.LogsService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class LogsService(object):
    """Service that can be used to push logs between one Application instrumented with
    OpenTelemetry and an collector, or between an collector and a central collector (in this
    case logs are sent/received to/from multiple Applications).
    """

    @staticmethod
    def Export(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/opentelemetry.proto.collector.logs.v1.LogsService/Export',
            opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceRequest.SerializeToString,
            opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from messagebus import TokenCreator, TokenPriority


MSGGRP_SLCTRL_PRODMGR = "9"

MSGGRP_SLCTRL_RUPTURE = "42"

TK_SLCTRL_PMGR_GET_TENDERS = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_PRODMGR,
    number="68",
)


TK_SLCTRL_PMGR_RUPTURE_UPDATE_ITEMS = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=MSGGRP_SLCTRL_RUPTURE,
    number="8",
)


def get_tokens():
    return (
        TK_SLCTRL_PMGR_GET_TENDERS,
    )



from typing import (
    Any,
    Dict,
)
from xml.etree import cElementTree as eTree

import iso8601
from common import FiscalOrderTotals
from nreports_api.shared import JSON
from ._BaseLoyaltyCoupon import BaseLoyaltyCoupon


class LoyaltyCouponNfceBuilder(BaseLoyaltyCoupon):

    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        pos_id = int(kwargs["posId"])
        order_id = int(kwargs["orderId"])
        fiscal_xml = kwargs["fiscalXml"]

        data_to_build["STORE_NAME"] = fiscal_xml.find(".//emit/xNome").text
        data_to_build["LOYALTY_QRCODE_URL"] = self._get_loyalty_qrcode_url(fiscal_xml=fiscal_xml)
        data_to_build["LOYALTY_PROTOCOL_DATE"] = self._get_protocol_date(fiscal_xml=fiscal_xml)

        self.fill_constant_values(
            pos_id=pos_id,
            order_id=order_id,
            data_to_build=data_to_build,
        )

        return data_to_build

    @staticmethod
    def _get_protocol_date(
        fiscal_xml: eTree.Element,
    ) -> str:

        protocol_date = iso8601.parse_date(fiscal_xml.find(".//dhEmi").text)

        return protocol_date.strftime("%d/%m/%Y %H:%M:%S")

    @staticmethod
    def _get_loyalty_qrcode_url(
        fiscal_xml: eTree.Element,
    ) -> str:

        fiscal_order_totals = FiscalOrderTotals(fiscal_xml=fiscal_xml)
        nfce_key = fiscal_xml.find(".//infNFe").attrib["Id"][3:]

        return f"{nfce_key}|{fiscal_order_totals.total_paid:.2f}"



from threading import (
    Lock,
    Condition,
    Thread,
)

from typing import (
    Dict,
    Optional,
)

from .. import logger
from actions.models import UserLockType


class UserLock(object):
    def __init__(
        self,               # type: UserLock
        user_id,            # type: str
        lock_reason,        # type: UserLockType
        lock_timeout,       # type: Optional[float]
    ):
        # type: (...) -> None

        self.user_id = user_id
        self.lock_reason = lock_reason
        self.lock_timeout = lock_timeout

        self._lock = Lock()
        self._lock.acquire()

        self.wait_condition = Condition()
        self.wait_thread = self._create_automatic_unlock_thread()
        self.wait_thread.start()

    def get_reason(self):
        return self.lock_reason

    def unlock(self):
        logger.debug("Unlock called for user: {}".format(self.user_id))
        with self.wait_condition:
            self.wait_condition.notify()

        self.wait_thread.join()

    def is_locked(self):
        return self._lock.locked()

    def _create_automatic_unlock_thread(self):
        thread = Thread(target=self._wait_and_unlock)
        thread.daemon = True
        return thread

    def _wait_and_unlock(self):
        self._wait_timeout()
        self._lock.release()

    def _wait_timeout(self):
        logger.debug("Waiting {} seconds to unlock {}".format(self.lock_timeout, self.user_id))
        with self.wait_condition:
            leave_without_timeout = self.wait_condition.wait(timeout=self.lock_timeout)

        if not leave_without_timeout:
            logger.debug("Timeout expired for user lock: {}".format(self.user_id))


class UserLocker(object):
    management_lock = Lock()

    def __init__(self):
        self.user_locks = {}  # type: Dict[str, UserLock]

    def lock_user(
        self,                    # type: UserLocker
        user_id,                 # type: str
        lock_reason,             # type: UserLockType
        lock_timeout=None,       # type: Optional[float]
    ):
        # type: (...) -> bool

        with UserLocker.management_lock:
            if self._is_locked(user_id=user_id):
                return False

            self.user_locks[user_id] = UserLock(
                user_id=user_id,
                lock_reason=lock_reason,
                lock_timeout=lock_timeout,
            )

        return True

    def unlock_user(self, user_id):
        with UserLocker.management_lock:
            if not self._is_locked(user_id=user_id):
                return

            self.user_locks[user_id].unlock()

    def get_lock_reason(
        self,           # type: UserLocker
        user_id,        # type: str
    ):
        # type: (...) -> Optional[UserLockType]

        with UserLocker.management_lock:
            if not self._is_locked(user_id=user_id):
                return None

            return self.user_locks[user_id].get_reason()

    def _is_locked(
        self,       # type: UserLocker
        user_id,    # type: str
    ):
        # type: (...) -> bool

        return user_id in self.user_locks and self.user_locks[user_id].is_locked()


user_locker = UserLocker()


def acquire_user_lock(
    user_id,                # type: str
    lock_reason,            # type: UserLockType
    lock_timeout=None,      # type: Optional[float]
):
    # type: (...) -> bool

    return user_locker.lock_user(
        user_id=user_id,
        lock_reason=lock_reason,
        lock_timeout=lock_timeout,
    )


def release_user_lock(
    user_id,  # type: str
):
    # type: (...) -> None

    user_locker.unlock_user(user_id=user_id)


def get_lock_reason(
    user_id,    # type: str
):
    # type: (...) -> Optional[UserLockType]

    return user_locker.get_lock_reason(user_id=user_id)



# gs1_128.py - functions for handling GS1-128 codes
#
# Copyright (C) 2019 Sergi Almacellas Abellana
# Copyright (C) 2020-2023 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""GS1-128 (Standard to encode product information in Code 128 barcodes).

The GS1-128 (also called EAN-128, UCC/EAN-128 or UCC-128) is an international
standard for embedding data such as best before dates, weights, etc. with
Application Identifiers (AI).

The GS1-128 standard is used as a product identification code on bar codes.
It embeds data with Application Identifiers (AI) that defines the kind of
data, the type and length. The standard is also known as UCC/EAN-128, UCC-128
and EAN-128.

GS1-128 is a subset of Code 128 symbology.

More information:

* https://en.wikipedia.org/wiki/GS1-128
* https://www.gs1.org/standards/barcodes/application-identifiers
* https://www.gs1.org/docs/barcodes/GS1_General_Specifications.pdf

>>> compact('(01)38425876095074(17)181119(37)1 ')
'013842587609507417181119371'
>>> encode({'01': '38425876095074'})
'0138425876095074'
>>> info('0138425876095074')
{'01': '38425876095074'}
>>> validate('(17)181119(01)38425876095074(37)1')
'013842587609507417181119371'
"""

import datetime
import decimal
import re

from stdnum import numdb
from stdnum.exceptions import *
from stdnum.util import clean


# our open copy of the application identifier database
_gs1_aidb = numdb.get('gs1_ai')


# Extra validation modules based on the application identifier
_ai_validators = {
    '01': 'stdnum.ean',
    '02': 'stdnum.ean',
    '8007': 'stdnum.iban',
}


def compact(number):
    """Convert the GS1-128 to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace. For a more consistent compact representation use
    :func:`validate()`.
    """
    return clean(number, '()').strip()


def _encode_value(fmt, _type, value):
    """Encode the specified value given the format and type."""
    if _type == 'decimal':
        if isinstance(value, (list, tuple)) and fmt.startswith('N3+'):
            number = _encode_value(fmt[3:], _type, value[1])
            return number[0] + value[0].rjust(3, '0') + number[1:]
        value = str(value)
        if fmt.startswith('N..'):
            length = int(fmt[3:])
            value = value[:length + 1]
            number, digits = (value.split('.') + [''])[:2]
            digits = digits[:9]
            return str(len(digits)) + number + digits
        else:
            length = int(fmt[1:])
            value = value[:length + 1]
            number, digits = (value.split('.') + [''])[:2]
            digits = digits[:9]
            return str(len(digits)) + (number + digits).rjust(length, '0')
    elif _type == 'date':
        if isinstance(value, (list, tuple)) and fmt == 'N6..12':
            return '%s%s' % (
                _encode_value('N6', _type, value[0]),
                _encode_value('N6', _type, value[1]))
        elif isinstance(value, datetime.date):
            if fmt in ('N6', 'N6..12'):
                return value.strftime('%y%m%d')
            elif fmt == 'N10':
                return value.strftime('%y%m%d%H%M')
            elif fmt in ('N6+N..4', 'N6[+N..4]'):
                value = value.strftime('%y%m%d%H%M')
                if value.endswith('00'):
                    value = value[:-2]
                if value.endswith('00'):
                    value = value[:-2]
                return value
            elif fmt in ('N8+N..4', 'N8[+N..4]'):
                value = value.strftime('%y%m%d%H%M%S')
                if value.endswith('00'):
                    value = value[:-2]
                if value.endswith('00'):
                    value = value[:-2]
                return value
            else:  # pragma: no cover (all formats should be covered)
                raise ValueError('unsupported format: %s' % fmt)
    return str(value)


def _max_length(fmt, _type):
    """Determine the maximum length based on the format ad type."""
    length = sum(int(re.match(r'^[NXY][0-9]*?[.]*([0-9]+)[\[\]]?$', x).group(1)) for x in fmt.split('+'))
    if _type == 'decimal':
        length += 1
    return length


def _pad_value(fmt, _type, value):
    """Pad the value to the maximum length for the format."""
    if _type in ('decimal', 'int'):
        return value.rjust(_max_length(fmt, _type), '0')
    return value.ljust(_max_length(fmt, _type))


def _decode_value(fmt, _type, value):
    """Decode the specified value given the fmt and type."""
    if _type == 'decimal':
        if fmt.startswith('N3+'):
            return (value[1:4], _decode_value(fmt[3:], _type, value[0] + value[4:]))
        digits = int(value[0])
        value = value[1:]
        if digits:
            value = value[:-digits] + '.' + value[-digits:]
        return decimal.Decimal(value)
    elif _type == 'date':
        if len(value) == 6:
            if value[4:] == '00':
                # When day == '00', it must be interpreted as last day of month
                date = datetime.datetime.strptime(value[:4], '%y%m')
                if date.month == 12:
                    date = date.replace(day=31)
                else:
                    date = date.replace(month=date.month + 1, day=1) - datetime.timedelta(days=1)
                return date.date()
            else:
                return datetime.datetime.strptime(value, '%y%m%d').date()
        elif len(value) == 12 and fmt in ('N12', 'N6..12'):
            return (_decode_value('N6', _type, value[:6]), _decode_value('N6', _type, value[6:]))
        else:
            # other lengths are interpreted as variable-length datetime values
            return datetime.datetime.strptime(value, '%y%m%d%H%M%S'[:len(value)])
    elif _type == 'int':
        return int(value)
    return value.strip()


def info(number, separator=''):
    """Return a dictionary containing the information from the GS1-128 code.

    The returned dictionary maps application identifiers to values with the
    appropriate type (`str`, `int`, `Decimal`, `datetime.date` or
    `datetime.datetime`).

    If a `separator` is provided it will be used as FNC1 to determine the end
    of variable-sized values.
    """
    number = compact(number)
    data = {}
    identifier = ''
    # skip separator
    if separator and number.startswith(separator):
        number = number[len(separator):]
    while number:
        # extract the application identifier
        ai, info = _gs1_aidb.info(number)[0]
        if not info or not number.startswith(ai):
            raise InvalidComponent()
        number = number[len(ai):]
        # figure out the value part
        value = number[:_max_length(info['format'], info['type'])]
        if separator and info.get('fnc1', False):
            idx = number.find(separator)
            if idx > 0:
                value = number[:idx]
        number = number[len(value):]
        # validate the value if we have a custom module for it
        if ai in _ai_validators:
            mod = __import__(_ai_validators[ai], globals(), locals(), ['validate'])
            mod.validate(value)
        # convert the number
        data[ai] = _decode_value(info['format'], info['type'], value)
        # skip separator
        if separator and number.startswith(separator):
            number = number[len(separator):]
    return data


def encode(data, separator='', parentheses=False):
    """Generate a GS1-128 for the application identifiers supplied.

    The provided dictionary is expected to map application identifiers to
    values. The supported value types and formats depend on the application
    identifier.

    If a `separator` is provided it will be used as FNC1 representation,
    otherwise variable-sized values will be expanded to their maximum size
    with appropriate padding.

    If `parentheses` is set the application identifiers will be surrounded
    by parentheses for readability.
    """
    ai_fmt = '(%s)' if parentheses else '%s'
    # we keep items sorted and keep fixed-sized values separate tot output
    # them first
    fixed_values = []
    variable_values = []
    for inputai, value in sorted(data.items()):
        ai, info = _gs1_aidb.info(str(inputai))[0]
        if not info:
            raise InvalidComponent()
        # validate the value if we have a custom module for it
        if ai in _ai_validators:
            mod = __import__(_ai_validators[ai], globals(), locals(), ['validate'])
            mod.validate(value)
        value = _encode_value(info['format'], info['type'], value)
        # store variable-sized values separate from fixed-size values
        if info.get('fnc1', False):
            variable_values.append((ai_fmt % ai, info['format'], info['type'], value))
        else:
            fixed_values.append(ai_fmt % ai + value)
    # we need the separator for all but the last variable-sized value
    # (or pad values if we don't have a separator)
    return ''.join(
        fixed_values + [
            ai + (value if separator else _pad_value(fmt, _type, value)) + separator
            for ai, fmt, _type, value in variable_values[:-1]
        ] + [
            ai + value
            for ai, fmt, _type, value in variable_values[-1:]
        ])


def validate(number, separator=''):
    """Check if the number provided is a valid GS1-128.

    This checks formatting of the number and values and returns a stable
    representation.

    If a separator is provided it will be used as FNC1 for both parsing the
    provided number and for encoding the returned number.
    """
    try:
        return encode(info(number, separator), separator)
    except ValidationError:
        raise
    except Exception:  # noqa: B902
        # We wrap all other exceptions to ensure that we only return
        # exceptions that are a subclass of ValidationError
        # (the info() and encode() functions expect some semblance of valid
        # input)
        raise InvalidFormat()


def is_valid(number, separator=''):
    """Check if the number provided is a valid GS1-128."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from typing import (
    Any,
    Dict,
    List,
)
from xml.etree import cElementTree as eTree

from common import FiscalOrderTotals
from common.builder_functions import get_optional_customer_doc
from model import FiscalPaymentTypes
from nreports.utils import get_logo_image_str
from nreports_api.shared import JSON
from .._BaseFiscalSat import BaseFiscalSat


class FiscalCouponSatBuilder(BaseFiscalSat):

    def build_data(
        self,
        data_to_build: JSON,
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        fiscal_xml = kwargs["fiscalXml"]

        fiscal_order_totals = FiscalOrderTotals(fiscal_xml=fiscal_xml)
        customer_document = get_optional_customer_doc(fiscal_xml=fiscal_xml)

        self._build_header(
            data_to_build=data_to_build,
            fiscal_xml=fiscal_xml,
        )
        data_to_build["STORE_IMG_LOGO"] = get_logo_image_str(image_path=self.images_path["STORE_IMG_LOGO"])

        self._build_body(
            data_to_build=data_to_build,
            fiscal_xml=fiscal_xml,
            fiscal_order_totals=fiscal_order_totals,
            customer_document=customer_document,
            products=self._get_item_products(fiscal_xml=fiscal_xml),
            tenders=self._get_item_tenders(fiscal_xml=fiscal_xml),
        )
        self._build_footer(
            data_to_build=data_to_build,
            fiscal_xml=fiscal_xml,
            customer_document=customer_document,
        )

        return data_to_build

    @staticmethod
    def _get_item_products(
        fiscal_xml: eTree.Element,
    ) -> List[Dict[str, str]]:

        item_products: List[Dict[str, str]] = []
        for item_product in fiscal_xml.findall(".//det/prod"):
            item_products.append(
                {
                    "ITEM_PRODUCT_CODE": item_product.find("cProd").text,
                    "ITEM_PRODUCT_NAME": item_product.find("xProd").text,
                    "ITEM_PRODUCT_QUANTITY": f"{float(item_product.find('qCom').text):.3f}",
                    "ITEM_PRODUCT_UNIT": item_product.find("uCom").text.upper(),
                    "ITEM_PRODUCT_PRICE": f"{float(item_product.find('vUnCom').text):.2f}",
                    "ITEM_PRODUCT_TOTAL_PRICE": f"{float(item_product.find('vProd').text):.2f}",
                }
            )

        return item_products

    @staticmethod
    def _get_item_tenders(
        fiscal_xml: eTree.Element,
    ) -> List[Dict[str, str]]:

        tender_types_by_id = {x[0]: x[1] for x in FiscalPaymentTypes.list_values()}
        tenders: List[Dict[str, str]] = []
        for item_product in fiscal_xml.findall(".//pgto/MP"):
            tender_name_type = tender_types_by_id.get(
                item_product.find("cMP").text,
                FiscalPaymentTypes.OUTROS.value[1],
            )

            tenders.append({
                "TENDER_TYPE": tender_name_type.upper(),
                "TENDER_VALUE": item_product.find("vMP").text,
            })

        return tenders



# -*- coding: utf-8 -*-
import json
import logging
from xml.etree.ElementTree import Element

import persistence
import pyscripts
import re
from datetime import (
    datetime,
    timedelta,
)
from xml.etree import cElementTree as eTree

from bustoken import (
    TK_ELECTRONIC_PAYMENT_GET_DOCUMENT,
    TK_GET_USED_SERVICE,
)
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
)
from posot import OrderTaker
from sysactions import (
    show_keyboard,
    show_messagebox,
    get_posot,
    get_model,
    get_current_order,
    translate_message,
    get_cfg,
    StopAction,
)

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from mw_helper import (
    format_currency,
    show_listbox_with_disabled_options,
    show_message_dialog,
)

from api.product import get_product_service
from application.domain import SaleLine
from model.customexception import ErrorGettingDevice

from pos_model import ProductType

mbcontext = pyscripts.mbcontext
logger = logging.getLogger("TableActions")

cached_tender_types = None
card_brands_cache = None
customer_info_config = {}

NOT_ELECTRONIC_TYPE_ID = (0, 9)


def update_custom_properties(
    customer_doc=None,             # type: Optional[str]
    customer_name=None,            # type: Optional[str]
):
    # type: (...) -> Dict[str, str]

    dict_sale = {}

    if customer_doc is not None:
        dict_sale.update({"CUSTOMER_DOC": customer_doc})
    if customer_name is not None:
        dict_sale.update({"CUSTOMER_NAME": customer_name})

    return dict_sale


def parse_loyalty_customer_info(
    loyalty_customer_info=None,     # type: Optional[str]
):
    # type: (...) -> Dict[str, str]

    customer_info = json.loads(loyalty_customer_info or "{}")
    customer_benefits = customer_info.pop("benefits", [])
    customer = customer_info.get("customerInfo", {}).get("customer")
    customer["isValidDocument"] = validate_document(customer_doc=customer.get("document", "")) is not False
    customer_info.get("customerInfo", {}).update(customer)

    result = dict(LOYALTY_CUSTOMER_INFO=customer_info)

    if customer_benefits:
        result["LOYALTY_CUSTOMER_BENEFITS"] = customer_benefits
        benefits_on_total = [
            benefit.get("id") for benefit in customer_benefits
            if benefit.get("automatic") == "ON_TOTAL"
        ]
        if benefits_on_total:
            result["LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"] = benefits_on_total

    return result


def get_customer_doc_after_paid(
    pos_id,                     # type: str
    pos_ot,                     # type: OrderTaker
    ask_document_pinpad=False,  # type: Optional[bool]
):
    # type: (...) -> tuple

    document = get_custom_property_value(pos_id, "CUSTOMER_DOC")
    if document:
        return document, False

    valid = False
    customer_doc = None
    while not valid:
        if ask_document_pinpad:
            customer_doc = input_document_pinpad_dialog(pos_id)

        if customer_doc is None:
            customer_doc = show_keyboard(
                posid=pos_id,
                message="$ONLY_NUMBER",
                title="$ENTER_CPF_CNPJ",
                buttons="$OK|$NOT_INFORM",
                mask="CPF_CNPJ",
                numpad=True,
            )

        if customer_doc == "NOT_INFORM":
            customer_doc = ""

        if customer_doc == "":
            ret = show_messagebox(pos_id, "$CONFIRM_EMPTY_CPF_CNPJ", buttons="$YES|$NO")
            if ret == 1:
                raise StopAction()
            break
        customer_doc = str(customer_doc)
        valid = validate_document(customer_doc)
        if not valid:
            show_messagebox(pos_id, "$INVALID_CPF_CNPJ")
            customer_doc = None

    fill_additional_info(pos_ot, "CPF", customer_doc)
    return customer_doc, True


def electronic_payment_get_document(
    pos_id      # type: str
):
    # type: (...) -> Optional[str]

    component_name = get_device_service(
        pos_id=pos_id,
        device_type="eft",
    )
    if not component_name:
        return None

    try:
        result = mbcontext.MB_EasySendMessage(
            dest_name=component_name,
            token=TK_ELECTRONIC_PAYMENT_GET_DOCUMENT,
            data=pos_id
        )
        if result.token == TK_SYS_ACK:
            return result.data

        return None

    except (Exception,):
        logger.exception("Error ask documents from pinpad")
        return None


def input_document_pinpad_dialog(
    pos_id  # type: str
):
    # type: (...) -> Optional[str]

    option = show_messagebox(
        pos_id,
        message="$ASK_CPF_CNPJ",
        buttons="$CANCEL_WITHOUT_ICON|$NO_WITHOUT_ICON|$MANUAL|$PINPAD"
    )

    if option == 0:
        raise StopAction()
    elif option == 1:
        customer_doc = "NOT_INFORM"
    elif option == 2:
        customer_doc = None
    else:
        customer_doc = electronic_payment_get_document(pos_id)

    return customer_doc


def validate_document(customer_doc):
    valid = False
    if len(customer_doc) <= 11:
        valid = validar_cpf(customer_doc)
    elif len(customer_doc) > 11:
        valid = validar_cnpj(customer_doc)
    return valid


def fill_additional_info(pos_ot, properties, value):
    if pos_ot.additionalInfo:
        pos_ot.additionalInfo += "|"
    else:
        pos_ot.additionalInfo = ""
    pos_ot.additionalInfo += "{0}={1}".format(properties, value)


def validar_cpf(
    cpf,    # type: str
):
    # type: (...) -> bool

    cpf = "".join(re.findall(r"\d", str(cpf)))

    if (not cpf) or (len(cpf) < 11):
        return False

    inteiros = map(int, cpf)
    novo = inteiros[:9]

    if all(num == inteiros[0] for num in inteiros):
        return False

    while len(novo) < 11:
        r = sum([(len(novo) + 1 - i) * v for i, v in enumerate(novo)]) % 11

        if r > 1:
            f = 11 - r
        else:
            f = 0
        novo.append(f)

    if novo == inteiros:
        return cpf

    return False


def validar_cnpj(
    cnpj      # type: str
):
    # type: (...) -> bool

    cnpj = "".join(re.findall(r"\d", str(cnpj)))

    if (not cnpj) or (len(cnpj) < 14):
        return False

    inteiros = map(int, cnpj)
    novo = inteiros[:12]

    if all(num == inteiros[0] for num in inteiros):
        return False

    prod = [5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2]
    while len(novo) < 14:
        r = sum([x * y for (x, y) in zip(novo, prod)]) % 11
        if r > 1:
            f = 11 - r
        else:
            f = 0
        novo.append(f)
        prod.insert(0, 6)

    if novo == inteiros:
        return cnpj

    return False


def json_serialize(dictionary):
    # type: (Dict) -> str
    try:
        return json.dumps(dictionary, default=lambda o: getattr(o, "__dict__", str(o)))
    except ValueError:
        return dictionary


def compare_objects(obj1, obj2):
    return object_to_json(obj1) == object_to_json(obj2)


def object_to_json(obj):
    return json.dumps(obj, sort_keys=True, default=lambda o: getattr(o, "__dict__", str(o)))


def get_seconds_to_renew_cache(seconds_to_renew_cache):
    seconds_to_renew = 60 * 5
    try:
        seconds_to_renew = int(seconds_to_renew_cache)
    except TypeError:
        pass
    return seconds_to_renew


def reached_seconds_to_renew_cache(last_cache_time, seconds_to_renew_cache):
    return not last_cache_time or datetime.now() > last_cache_time + timedelta(seconds=seconds_to_renew_cache)


def get_custom_property_value(
    pos_id,                 # type: str
    custom_property,        # type: str
    order_id=None,          # type: Optional[str]
    order_picture=None,     # type: Optional[eTree.XML]
    default_value=None,     # type: Optional[str]
):
    # type: (...) -> Optional[str]

    model = get_model(pos_id)
    pos_ot = get_posot(model)

    if not order_picture:
        if not order_id:
            order_picture_by_pos_id = pos_ot.orderPicture(pos_id)
            if order_picture_by_pos_id:
                order_picture = eTree.XML(order_picture_by_pos_id)
        else:
            order_picture = eTree.XML(pos_ot.orderPicture(pos_id, order_id))

    if not order_picture:
        return None
    current_property = order_picture.find(".//CustomOrderProperties/OrderProperty[@key='{}']".format(custom_property))
    return default_value if current_property is None else current_property.get("value")


def mask_phone(
        phone_number  # type: str
):
    # type: (...) -> str

    phone_number_size = len(phone_number)

    if phone_number_size in (8, 9):
        return "{}-{}".format(
            phone_number[:-4],
            phone_number[-4:]
        )
    if phone_number_size in (10, 11):
        return "({}) {}-{}".format(
            phone_number[0:2],
            phone_number[2:-4],
            phone_number[-4:]
        )

    return phone_number


def is_valid_string(
    string,             # type: str
    match_string=None   # type: Optional[str]
):
    # type: (...) -> bool

    if not match_string:
        match_string = "^[_a-zA-Z0-9 ,.-]+$"
    return bool(re.match(match_string, string))


def get_card_brands():
    # type: () -> Dict

    global card_brands_cache
    if card_brands_cache is not None:
        return card_brands_cache

    card_brands = {}
    conn = None
    try:
        conn = persistence.Driver().open(
            mbcontext,
            service_name="FiscalPersistence"
        )
        cursor = conn.pselect("fiscal_getBandeiras")
        for row in cursor:
            card_brands.update({int(row.get_entry("Bandeira")): row.get_entry("Descricao")})
    except (Exception,):
        logger.exception("Error getting cards brands")
    finally:
        if conn:
            conn.close()

    card_brands_cache = card_brands
    return card_brands_cache


def _count_option_listbox(
    option_text,    # type: str
    value,          # type: float
    options_dict    # type: List[Dict]
):
    # type: (...) -> int

    count = 0
    for option in options_dict:
        if option_text == option["text"] and value == option["value"]:
            count += 1

    return count


def options_dict_to_str_list(
    model,      # type: eTree.XML
    options     # type: List[Dict]
):
    # type: (...) -> List[str]

    listbox = []
    for option in options:
        formatted_value = format_currency(
            model=model,
            value=option["value"],
        )
        count = _count_option_listbox(
            option["text"],
            option["value"],
            options
        )
        if count > 1:
            option_text = "{0}: {1} ({2})".format(
                option["text"],
                formatted_value,
                option["count"]
            )
        else:
            option_text = "{0}: {1}".format(
                option["text"],
                formatted_value
            )
        listbox.append(option_text)

    return listbox


def check_has_tef(
    tenders,    # type: eTree.XML
):
    # type: (...) -> bool

    for tender in tenders:
        if is_electronic_payment(tender):
            return True

    return False


def get_tenders_options_dict(
    pos_id,                         # type: str
    model,                          # type: eTree.XML
    tenders,                        # type: eTree.XML
    card_brands,                    # type: Dict[int, str]
    group_digital_payments=False,   # type: bool
    total=0.0,                      # type: Optional[float]
):
    # type: (...) -> Tuple[List[Dict], List]

    if total:
        all_tenders_option = {
            "text": translate_message(
                model,
                "ALL_TENDERS"
            ),
            "value": total,
            "count": 1
        }
        options_dict = [all_tenders_option]
        initial_tef_index = 1
    else:
        options_dict = []
        initial_tef_index = 0

    disabled_items_index = []
    tef_not_available = get_device_service(
        pos_id=pos_id,
        device_type="eft",
    ) is None

    if group_digital_payments and check_has_tef(tenders):
        digital_payments_option = {
            "text": translate_message(
                model,
                "ELECTRONIC_PAYMENTS"
            ),
            "value": 0,
            "count": 1,
            "tenders": []
        }
        options_dict.append(digital_payments_option)

    option_text = ""
    for tender_index, tender in enumerate(tenders):
        value = float(tender.get("tenderAmount"))
        if value == 0:
            continue

        is_tef = is_electronic_payment(tender)
        if is_tef and tef_not_available:
            disabled_items_index.append(str(tender_index + 1))

        if is_tef:
            if not group_digital_payments:
                option_text = _get_option_text(
                    tender,
                    card_brands
                )
            else:
                options_dict[initial_tef_index]["value"] += float(value)
                options_dict[initial_tef_index]["tenders"] = options_dict[initial_tef_index].get("tenders") or []
                options_dict[initial_tef_index]["tenders"].append(tender.get("tenderId"))
        else:
            option_text = tender.get("tenderDescr").encode("utf-8")

        if not is_tef or (is_tef and not group_digital_payments):
            count = _count_option_listbox(
                option_text,
                value,
                options_dict
            )
            new_option = {
                "text": option_text,
                "tender_index": tender_index,
                "value": value,
                "count": count + 1
            }
            options_dict.append(new_option)

    if len(disabled_items_index) > 0:
        disabled_items_index.insert(0, "0")

    return options_dict, disabled_items_index


def _get_option_text(
    tender,         # type: eTree.Element
    card_brands,    # type: Dict[int, str]
):
    detail = json.loads(tender.get("tenderDetail"))
    tender_description = tender.get("tenderDescr").encode("utf-8")
    brand_code = int(detail["Bandeira"])
    brand = card_brands[int(detail["Bandeira"])] if brand_code in card_brands else ""
    card_number = str(detail["CardNumber"][:4] if "None" not in detail["CardNumber"][:4] else None)

    if card_number in ["", None]:
        return "{0} / {1}".format(
            tender_description,
            brand
        )

    return "{0} / {1} / {2}".format(tender_description, card_number, brand)


def get_order_tender(
    order_picture,  # type: eTree.XML
    tender_id       # type: str
):
    # type: (...) -> Optional[eTree.XML]

    tenders = order_picture.findall(".//TenderHistory/Tender")
    for tender in tenders:
        if tender.get("tenderId") == tender_id:
            return tender

    return None


def is_electronic_payment(
    tender   # type: Element
):
    # type: (...) -> bool

    return _get_electronic_type_id(tender) not in NOT_ELECTRONIC_TYPE_ID


def _get_electronic_type_id(
    tender   # type: Element
):
    # type: (...) -> int

    try:
        if type(tender) is dict:
            return int(tender.get("electronicTypeId"))
    except TypeError:
        pass

    if "electronicType" in tender:
        return int(tender["electronicType"])

    return int(tender.get("eletronicType"))


def get_customer_info_config(pos_id):
    global customer_info_config

    if int(pos_id) not in customer_info_config:
        document = get_cfg(pos_id).find_values("CustomerInfoRequest/Document") or []
        name = get_cfg(pos_id).find_values("CustomerInfoRequest/Name") or []

        customer_info_config[int(pos_id)] = {"document": document, "name": name}

    return customer_info_config[int(pos_id)]


def _translate_array(pos_id, messages):
    # type: (str, List[str]) -> List[str]

    model = get_model(pos_id)
    result = []
    for index, message in enumerate(messages, start=1):
        result.append("{} - {}".format(index, translate_message(model, message)))

    return result


def get_tender_info(
    tender_id=None,     # type: Optional[int]
):
    # type: (...) -> Optional[Any[Union[Dict, object]]]

    global cached_tender_types

    if cached_tender_types is None:
        product_service = get_product_service()
        cached_tender_types = product_service.get_tenders_dict()

    if tender_id is None:
        return cached_tender_types

    return cached_tender_types[tender_id] if tender_id in cached_tender_types else None


def _get_substring_occurrence(
    text,       # type: str
    character,  # type: str
):
    # type: (...) -> str

    first_occurrence = text.find(character)
    return character * (text.rfind(character) - first_occurrence + 1 if first_occurrence > -1 else 0)


def pad_barcode(
    barcode,   # type: str
):
    # type: (...) -> str

    barcode = barcode.replace(" ", "")
    barcode_2 = _get_substring_occurrence(
        text=barcode,
        character="2",
    )
    barcode_x = _get_substring_occurrence(
        text=barcode,
        character="X",
    )
    barcode_y = _get_substring_occurrence(
        text=barcode,
        character="Y",
    )
    barcode_w = _get_substring_occurrence(
        text=barcode,
        character="W",
    )

    return "{}{}{}{}{}".format(
        barcode_2,
        barcode_x,
        "0" * (13 - (len(barcode_2) + len(barcode_x) + len(barcode_y) + len(barcode_w))),
        barcode_y,
        barcode_w
    )


def get_sector_name(
    order,  # type: eTree
):
    # type: (...) -> Optional[str]

    sector = order.find(".//CustomOrderProperties/OrderProperty[@key='{}']".format("SECTOR"))
    if sector is not None:
        return sector.get("value", None)

    table_id = order.find(".//CustomOrderProperties/OrderProperty[@key='{}']".format("TABLE_ID"))
    if table_id is not None and "TAB-" in table_id.get("value", None):
        return "TAB"

    return None


def get_brand_name(
    order,  # type: eTree
):
    # type: (...) -> Optional[str]

    brand = order.find(".//CustomOrderProperties/OrderProperty[@key='BRAND']")
    if brand is not None:
        return brand.get("value", None)

    return None


def get_device_service(
    pos_id,         # type: str
    device_type,    # type: str
    sector=None,    # type: Optional[str]
    brand=None,     # type: Optional[str]
):
    # type: (...) -> str

    request = json.dumps(
        {
            "posId": pos_id,
            "deviceType": device_type,
            "sector": sector,
            "brand": brand,
        }
    )

    msg = mbcontext.MB_EasySendMessage(
        dest_name="DeviceManager",
        token=TK_GET_USED_SERVICE,
        format=FM_PARAM,
        data=request,
    )

    if msg.token == TK_SYS_ACK:
        return msg.data

    exception = ErrorGettingDevice(
        pos_id=pos_id,
        device_type=device_type,
    )
    raise exception


def get_device_service_error_message(
    model,          # type: str
    exception,      # type: ErrorGettingDevice
):
    message_device_text = translate_message(model, "DEVICE_NOT_FOUND")
    message_device_type = translate_message(model, exception.device_type.upper())
    return "{}: {}".format(message_device_text, message_device_type)


def send_order_canceled_event(
    pos_id,       # type: str
    order_id="",  # type: str
):
    # type: (...) -> None

    try:
        if not order_id and pos_id:
            model = get_model(posid=pos_id)
            order_id = get_current_order(model=model).get("orderId")

        event_data = {
            "posId": pos_id,
            "orderId": order_id,
        }
        mbcontext.MB_EasyEvtSend(
            subject="ORDER_CANCELED",
            type="",
            xml=json.dumps(event_data),
        )
        logger.info("ORDER_CANCELED event sent successfully! order_id={0}".format(order_id))
    except (Exception,):
        logger.exception("Error sending ORDER_CANCELED event")


def parse_qr_tag(text):
    try:
        root = eTree.fromstring("<p>" + text + "</p>")
        message = "" if root.text is None else root.text.strip()
        qr_data = None
        tag = root.find("qr")
        if tag is not None:
            qr_data = "" if tag.text is None else tag.text.strip()
    except Exception:  # noqa
        logger.exception("error parsing qr tag for customer final message: %r", text)
        message = text
        qr_data = None
    return message, qr_data


def is_table_service(
    model,  # type: eTree.Element
):
    # type: (...) -> bool

    order_properties = model.findall(".//CustomOrderProperties/OrderProperty")
    for order_property in order_properties:
        if order_property.get("key") == "TABLE_ID":
            return True

    return False


def ask_pager_number(
    pos_id,  # type: str
):
    # type: (...) -> str

    valid = False
    pager_number = None
    while not valid:
        message = "$ONLY_NUMBERS"
        title = "$ENTER_PAGER_NUMBER"
        pager_number = show_keyboard(
            posid=pos_id,
            message=message,
            title=title,
            buttons="$OK|$NOT_INFORM",
            numpad=True,
        )
        if pager_number == "NOT_INFORM":
            pager_number = None

        valid = validate_pager_number(
            pager_number=pager_number,
        )
        if not valid:
            pager_number = None
            show_messagebox(
                posid=pos_id,
                message="$INVALID_PAGER_NUMBER",
            )

    return pager_number


def validate_pager_number(
    pager_number,  # type: str
):
    # type: (...) -> bool

    if pager_number == "":
        return False

    return True


def show_order_tenders_options(
    pos_id,                    # type: str
    listbox_options,           # type: List[str]
    is_disabled=False,         # type: Optional[bool]
    has_receipt_card=True,     # type: Optional[bool]
):
    # type: (...) -> Dict

    if is_disabled:
        buttons = "$CLEAR_TENDERS|$PREVIOUS"
    else:
        disable_reprint = "" if has_receipt_card else "#DISABLED"
        buttons = "$REPRINT_RECEIPT{}|$CLEAR_TENDERS|$PREVIOUS".format(disable_reprint)

    selected_options = show_listbox_with_disabled_options(
        pos_id,
        listbox_options,
        buttons=buttons,
        message="$SELECT_TENDER",
    )

    return selected_options


def is_stored_order(
    model,          # type: eTree.Element
    pos_id=None,    # type: int
    order_id=None,  # type: int
):
    # type (...) -> bool

    order_picture = model.find("CurrentOrder/Order")
    if order_picture is None and order_id:
        pos_ot = get_posot(model)
        order_picture = eTree.XML(pos_ot.orderPicture(
            posid=pos_id,
            orderid=order_id,
        ))

    order_states = (order_picture.findall(".//StateHistory/State") or []) if order_picture is not None else []
    return any([True for state in order_states if state.get("state") == "STORED"])


def is_order_in_progress(
    model,  # type:  eTree.Element
):
    # type (...) -> bool

    current_order = model.find("CurrentOrder/Order")
    if current_order is None:
        return False

    return current_order.get("state") == "IN_PROGRESS"


def tender_has_receipt_card(
    tenders,    # type:  eTree
):
    # type: (...) -> bool

    for tender in tenders:
        tender_detail = _get_receipt_to_print(
            tender=tender,
        )
        if tender_detail:
            return True

    return False


def _get_receipt_to_print(
    tender,     # type: eTree
):
    # type: (...) -> Optional[Dict]

    if int(tender.get("eletronicType")) in (0, 9):
        return None

    tender_detail = json.loads(tender.get("tenderDetail"))
    receipt_customer = tender_detail.get("ReceiptCustomer", None)
    amount = tender_detail.get("tenderAmount", 0)

    if receipt_customer not in ["", None] and amount > 0:
        return tender_detail


def build_dict(
    key,               # type: str
    value,             # type: str
    dict_data=None,    # type: Optional[Dict]
):
    # type: (...) -> Dict

    if dict_data is None:
        dict_data = {}

    if value is not None:
        dict_data[key] = value
    return dict_data


def get_translated_exception_message(
    model,              # type: eTree.Element
    exception_message,  # type: str
):
    # type: (...) -> str

    message_to_translate = exception_message.split("$")
    if len(message_to_translate) > 1:
        message_to_translate = message_to_translate[-1]
    message = translate_message(model, message_to_translate)
    return message


def build_clean_customer_props_dict():
    # type: (...) -> dict

    return {
        "CUSTOMER_DOC": "clear",
        "CUSTOMER_NAME": "clear",
        "CUSTOMER_PHONE": "clear",
        "ADDRESS_REFERENCE": "clear",
        "CITY": "clear",
        "POSTAL_CODE": "clear",
        "COMPLEMENT": "clear",
        "NEIGHBORHOOD": "clear",
        "STATE": "clear",
        "STREET_NAME": "clear",
        "STREET_NUMBER": "clear",
    }


def get_pos_id_from_session(
    session_id,     # type: str
):
    # type: (...) -> str

    for data in session_id.split(","):
        if data.startswith("pos="):
            return data.split("=")[1]

    return ""


def get_user_id_from_session(
    session_id,     # type: str
):
    # type: (...) -> str

    for data in session_id.split(","):
        if data.startswith("user="):
            return data.split("=")[1]

    return ""


def get_session_id_info(
    session_id,         # type: str
):
    # type: (...) -> Dict[str, str]

    session_id_info = {}
    for session_data in session_id.split(","):
        key, value = session_data.split("=")
        session_id_info[key] = value

    return session_id_info


def get_sale_line_to_tag(
    model,      # type: eTree.Element
    sale_line,  # type: Dict[str, Any]
):
    # type: (...) -> Dict[str, Any]

    if sale_line["itemType"] == ProductType.PRODUCT:
        return _get_sale_line_to_tag_by_product(
            model=model,
            sale_line=sale_line,
        )

    if sale_line["itemType"] == ProductType.OPTION:
        return _get_sale_line_to_tag_by_option(
            model=model,
            sale_line=sale_line,
        )

    if sale_line["itemType"] == ProductType.CANADD:
        return _get_sale_line_to_tag_by_can_add(
            model=model,
            sale_line=sale_line,
        )

    return sale_line


def get_benefit_total_value(
    benefit_id,         # type: str
    order_picture,      # type: eTree.XML
):
    # type: (...) -> float

    benefit_list_value = order_picture.find(".//CustomOrderProperties/OrderProperty[@key='BENEFIT_LIST']").get("value")
    benefit_list_value_json = json.loads(benefit_list_value)
    benefit_list = json.loads(benefit_list_value_json.get("benefitList"))
    total_value = 0
    for benefit in benefit_list:
        if benefit_id == benefit.get("benefit").get("id"):
            total_value = benefit.get("benefit_total_value")

    return total_value


def value_is_close_to(
    value,              # type: float
    close_to,           # type: float
    precision=1e-3,     # type: float
):
    # type: (...) -> bool

    return abs(value - close_to) < precision


def _get_parent(
    sale_line,  # type: Dict[str, Any]
):
    # type: (...) -> Optional[SaleLine]

    return _get_ancestor(
        sale_line=sale_line,
        index=-1,
    )


def _get_grand_parent(
    sale_line,  # type: Dict[str, Any]
):
    # type: (...) -> Optional[SaleLine]

    return _get_ancestor(
        sale_line=sale_line,
        index=-2,
    )


def _get_ancestor(
    sale_line,  # type: Dict[str, Any]
    index,      # type: int
):
    # type: (...) -> Optional[SaleLine]

    min_size = (-index) + 1
    divided_context = sale_line["itemId"].split(".")
    if len(divided_context) < min_size:
        return None

    ancestor_part_code = divided_context[index]
    ancestor_context = ".".join([s for s in divided_context[:index]])

    return SaleLine(
        seat_number=None,
        order_id=None,
        line_number=sale_line["lineNumber"],
        item_id=ancestor_context,
        level=None,
        part_code=ancestor_part_code,
    )


def _is_same_sale_line(
    sale_line,      # type: SaleLine
    sale_line_xml,  # type: eTree.Element
):
    # type: (...) -> bool

    line_number = sale_line_xml.get("lineNumber")
    context = sale_line_xml.get("itemId")
    part_code = sale_line_xml.get("partCode")

    base_line_number = sale_line.line_number
    base_context = sale_line.item_id
    base_part_code = sale_line.part_code

    return line_number == base_line_number and context == base_context and part_code == base_part_code


def _get_sale_line_to_tag_by_product(
    model,      # type: eTree.Element
    sale_line,  # type: Dict[str, Any]
):
    # type: (...) -> Dict[str, Any]

    parent_sale_line = _get_parent(sale_line=sale_line)
    if not parent_sale_line:
        return sale_line

    order = model.find("CurrentOrder/Order")
    sale_lines = order.findall("SaleLine")

    for current_sale_line in sale_lines:
        is_same_line = _is_same_sale_line(
            sale_line=parent_sale_line,
            sale_line_xml=current_sale_line,
        )
        if not is_same_line:
            continue

        item_type = current_sale_line.get("itemType")
        if item_type == ProductType.PRODUCT:
            return dict(current_sale_line.items())

    return sale_line


def _get_sale_line_to_tag_by_can_add(
    model,      # type: eTree.Element
    sale_line,  # type: Dict[str, Any]
):
    # type: (...) -> Dict[str, Any]

    grand_parent_sale_line = _get_grand_parent(sale_line=sale_line)
    if not grand_parent_sale_line:
        return sale_line

    order = model.find("CurrentOrder/Order")
    sale_lines = order.findall("SaleLine")

    for current_sale_line in sale_lines:
        is_same_line = _is_same_sale_line(
            sale_line=grand_parent_sale_line,
            sale_line_xml=current_sale_line,
        )
        if not is_same_line:
            continue

        item_type = current_sale_line.get("itemType")

        if item_type == ProductType.CANADD:
            return _get_sale_line_to_tag_by_can_add(
                model=model,
                sale_line=dict(current_sale_line.items()),
            )

        if item_type == ProductType.COMBO:
            return sale_line

        if item_type == ProductType.PRODUCT:
            return dict(current_sale_line.items())

    return sale_line


def _get_sale_line_to_tag_by_option(
    model,      # type: eTree.Element
    sale_line,  # type: Dict[str, Any]
):
    # type: (...) -> Dict[str, Any]

    parent_sale_line = _get_parent(sale_line=sale_line)
    if not parent_sale_line:
        return sale_line

    order = model.find("CurrentOrder/Order")
    sale_lines = order.findall("SaleLine")
    for current_sale_line in sale_lines:
        is_same_line = _is_same_sale_line(
            sale_line=parent_sale_line,
            sale_line_xml=current_sale_line,
        )
        if not is_same_line:
            continue

        item_type = current_sale_line.get("itemType")
        if item_type == ProductType.CANADD:
            return _get_sale_line_to_tag_by_can_add(
                model=model,
                sale_line=dict(current_sale_line.items()),
            )

        if item_type in [ProductType.PRODUCT, ProductType.COMBO]:
            return dict(current_sale_line.items())

    return sale_line


def show_applied_discount_and_pending_value_to_pay(
    pos_id,                 # type: str
    model,                  # type: eTree.Element
    discount_amount,        # type: float
    pending_value_to_pay,   # type: Union[str, float]
    tender_type,            # type: str
):
    # type: (...) -> None

    discount_amount_formatted = format_currency(
        model=model,
        value=discount_amount,
    )
    new_amount_formatted = format_currency(
        model=model,
        value=float(pending_value_to_pay),
    )

    message = "$DISCOUNT_APPLIED|{}|{}|{}".format(discount_amount_formatted, tender_type, new_amount_formatted)
    show_message_dialog(
        pos_id=pos_id,
        message=message,
    )



from typing import (
    List,
    Optional,
)

from ._DeviceStatusHistory import DeviceStatusHistory


class DeviceStatus(object):
    def __init__(
        self,
        device_name,   # type: str
        enabled,       # type: bool
        history=None,  # type: Optional[List[DeviceStatusHistory]]
    ):
        # type: (...) -> None

        self.device_name = device_name
        self.enabled = enabled
        self.history = history if history is not None else []



from .order import (
    get_order_total,
)
from .repository import (
    get_configured_alerts_by_type,
    insert_event_into_repository,
)
from .store import (
    get_store_id,
    get_boh_url,
    get_boh_api_key,
)
from .table import (
    get_table_picture,
)



import json
import logging

from logging import Logger

from bustoken import (
    TK_NCOUPON_REMOVE_BENEFIT,
    TK_NCOUPON_TEST_CONNECTION,
)
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
    MBMessage,
)
from sysactions import send_message
from typing import List


class NCouponInteractor(object):
    def __init__(self):
        # type: () -> None

        self.logger = logging.getLogger("PosActions")   # type: Logger

    def test_connection(
        self,
        pos_id,     # type: bool
    ):
        # type: (...) -> bool

        try:
            data = json.dumps({
                "posId": pos_id
            })
            response = send_message("nCoupon", TK_NCOUPON_TEST_CONNECTION, FM_PARAM, data, timeout=30000000)
            return response.token == TK_SYS_ACK
        except Exception as _:
            self.logger.exception("Error testing nCoupon connection")
            return False

    @staticmethod
    def remove_benefits(
        pos_id,             # type: str
        benefit_keys=None,  # type: List[str]
    ):
        # type: (...) -> MBMessage

        data = json.dumps(dict({"posId": pos_id, "benefitKeys": benefit_keys}))
        return send_message(
            dest_name="nCoupon",
            token=TK_NCOUPON_REMOVE_BENEFIT,
            format=FM_PARAM,
            data=data,
            timeout=30000000,
        )



# nie.py - functions for handling Spanish foreigner identity codes
# coding: utf-8
#
# Copyright (C) 2012-2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NIE (Número de Identificación de Extranjero, Spanish foreigner number).

The NIE is an identification number for foreigners. It is a 9 digit number
where the first digit is either X, Y or Z and last digit is a checksum
letter.

More information:

* https://es.wikipedia.org/wiki/N%C3%BAmero_de_identidad_de_extranjero

>>> validate('x-2482300w')
'X2482300W'
>>> validate('x-2482300a')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('X2482300')  # digit missing
Traceback (most recent call last):
    ...
InvalidLength: ...
"""

from stdnum.es import dni
from stdnum.exceptions import *
from stdnum.util import isdigits


__all__ = ['compact', 'calc_check_digit', 'validate', 'is_valid']


# use the same compact function as DNI
compact = dni.compact


def calc_check_digit(number):
    """Calculate the check digit. The number passed should not have the
    check digit included."""
    # replace XYZ with 012
    number = str('XYZ'.index(number[0])) + number[1:]
    return dni.calc_check_digit(number)


def validate(number):
    """Check if the number provided is a valid NIE. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if not isdigits(number[1:-1]) or number[:1] not in 'XYZ':
        raise InvalidFormat()
    if len(number) != 9:
        raise InvalidLength()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is a valid NIE. This checks the length,
    formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



import string
import warnings
from json import loads

from jmespath.exceptions import LexerError, EmptyExpressionError


class Lexer(object):
    START_IDENTIFIER = set(string.ascii_letters + '_')
    VALID_IDENTIFIER = set(string.ascii_letters + string.digits + '_')
    VALID_NUMBER = set(string.digits)
    WHITESPACE = set(" \t\n\r")
    SIMPLE_TOKENS = {
        '.': 'dot',
        '*': 'star',
        ']': 'rbracket',
        ',': 'comma',
        ':': 'colon',
        '@': 'current',
        '(': 'lparen',
        ')': 'rparen',
        '{': 'lbrace',
        '}': 'rbrace',
    }

    def tokenize(self, expression):
        self._initialize_for_expression(expression)
        while self._current is not None:
            if self._current in self.SIMPLE_TOKENS:
                yield {'type': self.SIMPLE_TOKENS[self._current],
                       'value': self._current,
                       'start': self._position, 'end': self._position + 1}
                self._next()
            elif self._current in self.START_IDENTIFIER:
                start = self._position
                buff = self._current
                while self._next() in self.VALID_IDENTIFIER:
                    buff += self._current
                yield {'type': 'unquoted_identifier', 'value': buff,
                       'start': start, 'end': start + len(buff)}
            elif self._current in self.WHITESPACE:
                self._next()
            elif self._current == '[':
                start = self._position
                next_char = self._next()
                if next_char == ']':
                    self._next()
                    yield {'type': 'flatten', 'value': '[]',
                           'start': start, 'end': start + 2}
                elif next_char == '?':
                    self._next()
                    yield {'type': 'filter', 'value': '[?',
                           'start': start, 'end': start + 2}
                else:
                    yield {'type': 'lbracket', 'value': '[',
                           'start': start, 'end': start + 1}
            elif self._current == "'":
                yield self._consume_raw_string_literal()
            elif self._current == '|':
                yield self._match_or_else('|', 'or', 'pipe')
            elif self._current == '&':
                yield self._match_or_else('&', 'and', 'expref')
            elif self._current == '`':
                yield self._consume_literal()
            elif self._current in self.VALID_NUMBER:
                start = self._position
                buff = self._consume_number()
                yield {'type': 'number', 'value': int(buff),
                       'start': start, 'end': start + len(buff)}
            elif self._current == '-':
                # Negative number.
                start = self._position
                buff = self._consume_number()
                if len(buff) > 1:
                    yield {'type': 'number', 'value': int(buff),
                           'start': start, 'end': start + len(buff)}
                else:
                    raise LexerError(lexer_position=start,
                                     lexer_value=buff,
                                     message="Unknown token '%s'" % buff)
            elif self._current == '"':
                yield self._consume_quoted_identifier()
            elif self._current == '<':
                yield self._match_or_else('=', 'lte', 'lt')
            elif self._current == '>':
                yield self._match_or_else('=', 'gte', 'gt')
            elif self._current == '!':
                yield self._match_or_else('=', 'ne', 'not')
            elif self._current == '=':
                if self._next() == '=':
                    yield {'type': 'eq', 'value': '==',
                        'start': self._position - 1, 'end': self._position}
                    self._next()
                else:
                    if self._current is None:
                        # If we're at the EOF, we never advanced
                        # the position so we don't need to rewind
                        # it back one location.
                        position = self._position
                    else:
                        position = self._position - 1
                    raise LexerError(
                        lexer_position=position,
                        lexer_value='=',
                        message="Unknown token '='")
            else:
                raise LexerError(lexer_position=self._position,
                                 lexer_value=self._current,
                                 message="Unknown token %s" % self._current)
        yield {'type': 'eof', 'value': '',
               'start': self._length, 'end': self._length}

    def _consume_number(self):
        start = self._position
        buff = self._current
        while self._next() in self.VALID_NUMBER:
            buff += self._current
        return buff

    def _initialize_for_expression(self, expression):
        if not expression:
            raise EmptyExpressionError()
        self._position = 0
        self._expression = expression
        self._chars = list(self._expression)
        self._current = self._chars[self._position]
        self._length = len(self._expression)

    def _next(self):
        if self._position == self._length - 1:
            self._current = None
        else:
            self._position += 1
            self._current = self._chars[self._position]
        return self._current

    def _consume_until(self, delimiter):
        # Consume until the delimiter is reached,
        # allowing for the delimiter to be escaped with "\".
        start = self._position
        buff = ''
        self._next()
        while self._current != delimiter:
            if self._current == '\\':
                buff += '\\'
                self._next()
            if self._current is None:
                # We're at the EOF.
                raise LexerError(lexer_position=start,
                                 lexer_value=self._expression[start:],
                                 message="Unclosed %s delimiter" % delimiter)
            buff += self._current
            self._next()
        # Skip the closing delimiter.
        self._next()
        return buff

    def _consume_literal(self):
        start = self._position
        lexeme = self._consume_until('`').replace('\\`', '`')
        try:
            # Assume it is valid JSON and attempt to parse.
            parsed_json = loads(lexeme)
        except ValueError:
            try:
                # Invalid JSON values should be converted to quoted
                # JSON strings during the JEP-12 deprecation period.
                parsed_json = loads('"%s"' % lexeme.lstrip())
                warnings.warn("deprecated string literal syntax",
                              PendingDeprecationWarning)
            except ValueError:
                raise LexerError(lexer_position=start,
                                 lexer_value=self._expression[start:],
                                 message="Bad token %s" % lexeme)
        token_len = self._position - start
        return {'type': 'literal', 'value': parsed_json,
                'start': start, 'end': token_len}

    def _consume_quoted_identifier(self):
        start = self._position
        lexeme = '"' + self._consume_until('"') + '"'
        try:
            token_len = self._position - start
            return {'type': 'quoted_identifier', 'value': loads(lexeme),
                    'start': start, 'end': token_len}
        except ValueError as e:
            error_message = str(e).split(':')[0]
            raise LexerError(lexer_position=start,
                             lexer_value=lexeme,
                             message=error_message)

    def _consume_raw_string_literal(self):
        start = self._position
        lexeme = self._consume_until("'").replace("\\'", "'")
        token_len = self._position - start
        return {'type': 'literal', 'value': lexeme,
                'start': start, 'end': token_len}

    def _match_or_else(self, expected, match_type, else_type):
        start = self._position
        current = self._current
        next_char = self._next()
        if next_char == expected:
            self._next()
            return {'type': match_type, 'value': current + next_char,
                    'start': start, 'end': start + 1}
        return {'type': else_type, 'value': current,
                'start': start, 'end': start}



from messagebus import DataType
from typing import (
    Any,
    Optional,
    Union,
)
from xml.etree import cElementTree as eTree

from msgbus import (
    TK_PRN_PRINT,
    MBTimeout,
)
from msgbus import (
    TK_SYS_NAK,
    MBException,
)
from .. import mb_context

from model.customexception import (
    ErrorGettingDevice,
    PrinterError,
)
from utilfunctions import (
    get_sector_name,
    get_device_service,
    get_device_service_error_message,
)
from sysactions import get_model


def send_json_data_print(
    pos_id,         # type: str
    print_data,     # type: str
    model=None,     # type: Optional[eTree.Element]
    order=None,     # type: Optional[eTree.Element]
):
    # type: (...) -> None

    try:
        current_printer = _get_printer(
            pos_id=pos_id,
            order=order,
        )
        msg = mb_context.MB_EasySendMessage(
            dest_name=current_printer,
            token=TK_PRN_PRINT,
            format=DataType.json,
            data=print_data,
            timeout=30 * 1000000,
        )
        if msg.token == TK_SYS_NAK:
            raise PrinterError(msg.data)

    except ErrorGettingDevice as ex:
        if model is None:
            model = get_model(posid=pos_id)

        message = get_device_service_error_message(
            model=model,
            exception=ex,
        )
        raise PrinterError(message)
    except (MBException, MBTimeout) as ex:
        raise PrinterError(ex)


def _get_printer(
    pos_id,         # type: Any[Union[str, int]]
    order=None,     # type: Optional[eTree.Element]
):
    # type: (...) -> str

    sector_name = get_sector_name(order=order) if order else None
    current_printer = get_device_service(
        pos_id=pos_id,
        device_type="printer",
        sector=sector_name,
    )
    return current_printer



# -*- coding: utf-8 -*-

from datetime import datetime

from application.model import (
    UpdateData,
    UpdateType,
    UpdateStatus,
)
from application.model.configuration import Configurations
from application.model.customexception import (
    ErrorInsertingNewUpdateVersion,
    ErrorGettingUsersList,
    ErrorUpdatingUserControlType,
)
from mbcontextmessagehandler import MbContextMessageBus
from msgbus import (
    TK_SYS_ACK,
    TK_POS_LISTUSERS,
    TK_POS_GETMODEL
)
from messagebus import Message, DataType
from mwhelper import (
    BaseRepository,
)
from persistence import Connection
from typing import (
    Any,
    Optional,
    Union
)
from xml.etree import cElementTree as etree

from ._CommonPOSRepository import CommonPOSRepository


class LoaderUpdaterPOSRepository(BaseRepository):
    def __init__(self, message_bus, configs):
        # type: (MbContextMessageBus, Configurations) -> None

        super(LoaderUpdaterPOSRepository, self).__init__(message_bus.mbcontext)

        self.message_bus = message_bus
        self.configs = configs
        self.update_type = UpdateType.loader
        self.update_value = self.update_type.value
        self.local_configs = configs.updaters[self.update_type.name]
        self.common_pos_repository = CommonPOSRepository(message_bus)

    def insert_new_loader_update(
        self,                   # type: LoaderUpdaterPOSRepository
        update_id,              # type: str
        update_name,            # type: str
        external_update_id,     # type: str
        obtained_date=None,     # type: Optional[str]
    ):
        # type: (...) -> None

        try:
            def inner_func(
                conn,   # type: Connection
            ):
                # type: (Connection) -> None

                if external_update_id:
                    formatted_query = self.get_update_by_external_update_id_query.format(
                        external_update_id,
                        self.update_value,
                    )
                    cursor = conn.select(formatted_query)
                    row = cursor.get_row(0)

                    if row and not obtained_date:
                        return

                external_update_id_value = "'{}'".format(external_update_id) if external_update_id else "NULL"
                if obtained_date is None:
                    new_obtained_date = "'{}'".format(datetime.now().isoformat())
                    download_and_backup_date = "'{}'".format(new_obtained_date) if not external_update_id else "NULL"
                else:
                    new_obtained_date = "'{}'".format(obtained_date)
                    download_and_backup_date = new_obtained_date

                query_format = self.update_query.format(
                    update_id,
                    update_name,
                    self.update_value,
                    UpdateStatus.pending,
                    new_obtained_date,
                    download_and_backup_date,
                    external_update_id_value,
                )
                conn.query(query_format)

            self.execute_with_connection(
                method=inner_func,
                service_name=self.configs.persistence_name,
            )
        except Exception as ex:
            raise ErrorInsertingNewUpdateVersion(ex)

    def insert_new_remote_loader_update(
        self,                   # type: LoaderUpdaterPOSRepository
        update_id,              # type: str
        update_name,            # type: str
        external_update_id,     # type: str
    ):
        # type: (...) -> bool

        try:
            def inner_func(
                conn,   # type: Connection
            ):
                # type: (Connection) -> bool

                formatted_query = self.get_update_by_external_update_id_query.format(
                    external_update_id,
                    self.update_value,
                )
                cursor = conn.select(formatted_query)
                row = cursor.get_row(0)

                if row:
                    return False

                external_update_id_value = "'{}'".format(external_update_id)
                date_now = "'{}'".format(datetime.now().isoformat())
                download_and_backup_date = "'{}'".format(date_now) if not external_update_id else "NULL"

                query_format = self.update_query.format(
                    update_id,
                    update_name,
                    self.update_value,
                    UpdateStatus.pending,
                    date_now,
                    download_and_backup_date,
                    external_update_id_value,
                )
                conn.query(query_format)

                return True

            return self.execute_with_connection(
                method=inner_func,
                service_name=self.configs.persistence_name,
            )
        except Exception as ex:
            raise ErrorInsertingNewUpdateVersion(ex)

    def get_last_loader_update_param(
        self,   # type: LoaderUpdaterPOSRepository
        param,  # type: str
    ):
        # type: (...) -> str

        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> Union[str, int]

            query = """
                SELECT *
                FROM UpdatesController
                WHERE TypeId = '{}'
                ORDER BY Id DESC
                LIMIT 1""".format(self.update_value)

            cursor = conn.select(query)
            row = cursor.get_row(0)

            if row:
                return row.get_entry(param)

            return 0

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def get_last_loader_update(
        self,   # type: LoaderUpdaterPOSRepository
    ):
        # type: (...) -> UpdateData

        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> Union[str, int]

            query = """
                SELECT *
                FROM UpdatesController
                WHERE TypeId = '{}'
                ORDER BY Id DESC
                LIMIT 1""".format(self.update_value)

            cursor = conn.select(query)
            row = cursor.get_row(0)

            return UpdateData(
                update_id=row.get_entry('UpdateId'),
                update_name=row.get_entry('UpdateName'),
                obtained_date=row.get_entry('ObtainedDate'),
                applied_date=row.get_entry('AppliedDate'),
                server_update_identifier=row.get_entry('ServerUpdateIdentifier'),
            )

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def update_pos_ctrl_type(self, pos_id, user_control_type):
        # type: (int, int) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                query = """
                    UPDATE PosState
                    SET UserControlType = '{0}'
                    WHERE PosId = '{1}'""".format(user_control_type, pos_id)
                conn.query(query)

            self.execute_with_connection(inner_func, service_name='Persistence')
        except Exception as ex:
            raise ErrorUpdatingUserControlType(ex)

    def get_pending_loader_update_id(self):
        # type: () -> int

        def inner_func(conn):
            # type: (Connection) -> int

            query = """
                SELECT UpdateId
                FROM UpdatesController
                WHERE TypeId = '{0}' AND StatusId = '{1}' AND AppliedDate IS NULL
                ORDER BY Id DESC""".format(self.update_value, UpdateStatus.pending)

            cursor = conn.select(query)
            row = cursor.get_row(0)

            if row:
                update_id = row.get_entry("UpdateId")
                return int(update_id)

            return 0

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def has_pending_loader_to_update(
        self,   # type: LoaderUpdaterPOSRepository
    ):
        # type: (...) -> bool

        def inner_func(
            conn,   # type: Connection
        ):
            # type: (...) -> bool

            query = """
                SELECT UpdateId
                FROM UpdatesController
                WHERE TypeId = '{}' AND AppliedDate IS NULL
                ORDER BY Id DESC
            """.format(self.update_value)

            cursor = conn.select(query)
            row = cursor.get_row(0)

            return row is not None

        return self.execute_with_connection(
            method=inner_func,
            service_name=self.configs.persistence_name,
        )

    def delete_update_line(
        self,                  # type: LoaderUpdaterPOSRepository
        external_update_id,    # type: str
        update_id,             # type: int
    ):
        # type: (...) -> None

        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> None

            query = """
                DELETE FROM UpdatesController
                WHERE ServerUpdateIdentifier = '{0}'
                AND UpdateId = '{1}'
                AND AppliedDate IS NULL;
            """.format(external_update_id, update_id)

            conn.select(query)

        self.execute_with_connection(
            method=inner_func,
            service_name=self.configs.persistence_name,
        )

    def update_loader_notified_date(
        self,                  # type: LoaderUpdaterPOSRepository
        update_id,             # type: int
    ):
        # type: (...) -> None

        def inner_func(
            conn,  # type: Connection
        ):
            # type: (...) -> None

            date_now = str(datetime.now().isoformat())
            query = """
                UPDATE UpdatesController
                SET NotifiedDate = '{0}'
                WHERE UpdateId <= '{1}' AND TypeId = 4;
            """.format(date_now, update_id)

            conn.select(query)

        self.execute_with_connection(
            method=inner_func,
            service_name=self.configs.persistence_name,
        )

    def mark_loader_update_as_applied(
            self,
    ):
        # type: (...) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                date_now = str(datetime.now().isoformat())
                query = """
                        UPDATE UpdatesController
                        SET StatusId = '{0}', AppliedDate = '{1}'
                        WHERE UpdateName = 'Loaders' AND AppliedDate IS NULL
                        """.format(UpdateStatus.applied, date_now)

                conn.query(query)

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorUpdatingUserControlType(ex)

    def mark_loader_update_as_error(
            self,
    ):
        # type: (...) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                query = """
                        UPDATE UpdatesController
                        SET StatusId = '{0}'
                        WHERE UpdateName = 'Loaders' AND AppliedDate IS NULL
                        """.format(UpdateStatus.error)

                conn.query(query)

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorUpdatingUserControlType(ex)

    def has_already_applied_update(
        self,   # type: LoaderUpdaterPOSRepository
    ):
        # type: (...) -> bool

        return self.common_pos_repository.has_already_applied_update(
            persistence_name=self.configs.persistence_name,
            update_value=self.update_value,
        )

    def get_last_loader_update_date(
        self,   # type: LoaderUpdaterPOSRepository
    ):
        # type: (...) -> Optional[datetime]

        return self.common_pos_repository.get_last_update_date(
            persistence_name=self.configs.persistence_name,
            update_value=self.update_value,
        )

    def get_pos_users(
        self,       # type: LoaderUpdaterPOSRepository
        pos_id,     # type: str
    ):
        # type: (...) -> str

        message = Message(
            token=TK_POS_LISTUSERS,
            data_type=DataType.param,
            data="{}".format(pos_id),
        )
        res = self.message_bus.send_message(
            component="POS{}".format(pos_id),
            message=message,
        )
        if res.token != TK_SYS_ACK:
            raise ErrorGettingUsersList("Could not list users in POS{}".format(pos_id))
        return res.data

    def has_pending_loader_update(
        self,   # type: LoaderUpdaterPOSRepository
    ):
        # type: (...) -> bool

        try:
            def inner_func(
                conn,   # type: Connection
            ):
                # type: (...) -> bool

                query = """
                    SELECT
                        NotifiedDate
                    FROM
                        UpdatesController
                    WHERE
                        TypeId = {0}
                        AND NotifiedDate IS NULL
                        AND AppliedDate IS NOT NULL
                        AND UpdateId = (
                            SELECT
                                MAX(UpdateId)
                            FROM
                                UpdatesController
                            WHERE
                                TypeId = {0}
                        )
                """.format(self.update_value)

                cursor = conn.select(query)
                row = cursor.get_row(0)
                if row:
                    return True
                return False

            return self.execute_with_connection(
                method=inner_func,
                service_name=self.configs.persistence_name,
            )
        except Exception as ex:
            raise ErrorUpdatingUserControlType(ex)

    def get_model(
        self,
        pos_id    # type: Any[Union[str, int]]
    ):
        # type: (...) -> etree.XML

        message = Message(TK_POS_GETMODEL, data_type=DataType.param, data="{}".format(pos_id))
        res = self.message_bus.send_message("POS{}".format(pos_id), message)
        if res.token != TK_SYS_ACK:
            raise ErrorGettingUsersList("Could not list users in POS{}".format(pos_id))
        return etree.XML(res.data)

    update_query = """
        INSERT INTO UpdatesController
        (
            UpdateId,
            UpdateName,
            TypeId,
            StatusId,
            ObtainedDate,
            DownloadedDate,
            BackupDate,
            ServerUpdateIdentifier
        )
        VALUES ('{0}', '{1}', '{2}', '{3}', {4}, {5}, {5}, {6})
    """

    get_update_by_external_update_id_query = """
        SELECT ServerUpdateIdentifier
        FROM UpdatesController
        WHERE ServerUpdateIdentifier = '{0}'
        AND TypeId = {1}
    """



from xml.etree.ElementTree import Element

from xml.etree import cElementTree as eTree

from messagebus import (
    MessageBus,
    Message,
    DataType,
    DefaultToken,
)
from typing import (
    Optional,
)
from .token import (
    TK_POS_GET_MODEL,
)


def requires_model(f):
    def wrapper(*args, **kwargs):
        if args[0].model is None:
            args[0].model = args[0]._retrieve_new_model()

        return f(*args, **kwargs)

    return wrapper


class PosService(object):

    def __init__(
        self,               # type: PosService
        message_bus,        # type: MessageBus
        pos_id,             # type: int
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.pos_id = pos_id

        self.component_name = "POS{}".format(self.pos_id)

        self.model = None

    def get_model(self):
        # type: (...) -> Element

        if self.model is None:
            self.model = self._retrieve_new_model()

        return self.model

    def get_updated_model(self):
        # type: (...) -> Element

        self.model = self._retrieve_new_model()
        return self.model

    @requires_model
    def get_operator_session(self):
        # type: (...) -> Optional[str]

        for op in self.model.findall("Operator"):
            if op.get("state") == "LOGGEDIN":
                return op.get("sessionId")

        return None

    @requires_model
    def get_pod_type(self):
        # type: (...) -> str

        return self.model.find("WorkingMode").get("podType")

    @requires_model
    def get_pos_function(self):
        # type: (...) -> str

        return self.model.find("WorkingMode").get("posFunction")

    def _retrieve_new_model(self):
        # type: (...) -> Element

        msg = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=TK_POS_GET_MODEL,
                data_type=DataType.string,
                data=str(self.pos_id).encode()
            )
        )
        if msg.token != DefaultToken.TK_SYS_ACK:
            raise RuntimeError("Could not retrieve state of pos '{}'".format(self.pos_id))

        return eTree.XML(msg.data)



# AST nodes have this structure:
# {"type": <node type>", children: [], "value": ""}


def comparator(name, first, second):
    return {'type': 'comparator', 'children': [first, second], 'value': name}


def current_node():
    return {'type': 'current', 'children': []}


def expref(expression):
    return {'type': 'expref', 'children': [expression]}


def function_expression(name, args):
    return {'type': 'function_expression', 'children': args, 'value': name}


def field(name):
    return {"type": "field", "children": [], "value": name}


def filter_projection(left, right, comparator):
    return {'type': 'filter_projection', 'children': [left, right, comparator]}


def flatten(node):
    return {'type': 'flatten', 'children': [node]}


def identity():
    return {"type": "identity", 'children': []}


def index(index):
    return {"type": "index", "value": index, "children": []}


def index_expression(children):
    return {"type": "index_expression", 'children': children}


def key_val_pair(key_name, node):
    return {"type": "key_val_pair", 'children': [node], "value": key_name}


def literal(literal_value):
    return {'type': 'literal', 'value': literal_value, 'children': []}


def multi_select_dict(nodes):
    return {"type": "multi_select_dict", "children": nodes}


def multi_select_list(nodes):
    return {"type": "multi_select_list", "children": nodes}


def or_expression(left, right):
    return {"type": "or_expression", "children": [left, right]}


def and_expression(left, right):
    return {"type": "and_expression", "children": [left, right]}


def not_expression(expr):
    return {"type": "not_expression", "children": [expr]}


def pipe(left, right):
    return {'type': 'pipe', 'children': [left, right]}


def projection(left, right):
    return {'type': 'projection', 'children': [left, right]}


def subexpression(children):
    return {"type": "subexpression", 'children': children}


def slice(start, end, step):
    return {"type": "slice", "children": [start, end, step]}


def value_projection(left, right):
    return {'type': 'value_projection', 'children': [left, right]}



from abc import ABCMeta, abstractmethod
from datetime import timedelta
from logging import Logger

import iso8601
from mw_helper import ensure_iterable, ensure_list
from production.box import ProductionBox
from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import (
    ProductionOrder,
    ProdStates,
)
from typing import Optional, Union, List


def _get_order_time(
    order,                  # type: ProductionOrder
    time_type="display",    # type: Optional[str]
):
    # type: (...) -> str

    selected_time = order.display_time
    if time_type == "in_preparation":
        is_delivered = order.prod_state == ProdStates.DELIVERED
        selected_time = order.prod_state_last_update if is_delivered else order.properties.get("FISCALIZATION_DATE", "")

    if selected_time == "":
        selected_time = order.created_at
    return selected_time


class OrderSequencerBox(OrderChangerProductionBox):
    def __init__(self, name, sons, sorters, logger=None):
        # type: (str, Optional[Union[List[ProductionBox]], ProductionBox], Optional[Union[List[OrderSorter]]], Logger) -> None # noqa
        super(OrderSequencerBox, self).__init__(name, sons, logger)
        self.sorters = ensure_iterable(sorters)  # type: List[OrderSequencerBox.OrderSorter]

    def change_order(self, order):
        sequence = ""
        for sorter in self.sorters:
            sequence += sorter.create_prod_sequence(order)

        order.prod_sequence = sequence
        return order

    class OrderSorter(object):
        __metaclass__ = ABCMeta

        @abstractmethod
        def create_prod_sequence(self, order):
            # type: (ProductionOrder) -> str
            raise NotImplementedError()

    class PriorityCustomerSorter(OrderSorter):
        def create_prod_sequence(self, order):
            if "PRIORITY" in order.properties and order.properties["PRIORITY"] == "true":
                return "00"
            for item in order.items:
                if 'highPriority' in item.properties:
                    return "10"
            return "99"

    class PodTypeSorter(OrderSorter):
        def __init__(self, pod_type_order):
            # type: (Optional[Union[List[str], str]]) -> None
            self.pod_type_order = ensure_list(pod_type_order)

        def create_prod_sequence(self, order):
            try:
                index = self.pod_type_order.index(order.pod_type)
                return str(index).zfill(2)
            except ValueError:
                return "99"

    class DoneInFrontSorter(OrderSorter):
        def create_prod_sequence(self, order):
            if order.tagged_timestamp is not None:
                return "00" + order.tagged_timestamp
            else:
                return "99"

    class SaleTypeSorter(OrderSorter):
        def __init__(self, sale_type_order):
            self.sale_type_order = ensure_list(sale_type_order)

        def create_prod_sequence(self, order):
            try:
                index = self.sale_type_order.index(order.sale_type)
                return str(index).zfill(2)
            except ValueError:
                return "99"

    class OrderIdSorter(OrderSorter):
        def create_prod_sequence(self, order):
            return str(order.order_id)

    class OrderStateSorter(OrderSorter):
        def __init__(self, state_order):
            self.state_order = ensure_list(state_order)

        def create_prod_sequence(self, order):
            try:
                index = self.state_order.index(order.state)
                return str(index).zfill(2)
            except ValueError:
                return "99"

    class LastModifiedSorter(OrderSorter):
        def __init__(self, last_modified_first=False):
            super(OrderSequencerBox.LastModifiedSorter, self).__init__()
            self.last_modified_first = last_modified_first

        def create_prod_sequence(self, order):
            if self.last_modified_first:
                future_date = iso8601.parse_date("3000-02-13T00:00:00Z")
                past_date = iso8601.parse_date("1984-02-13T00:00:00Z")
                last_change_date = iso8601.parse_date(order.state_history[-1].timestamp)
                return past_date + timedelta(seconds=(future_date - last_change_date).total_seconds())
            else:
                return order.state_history[-1].timestamp

    class DisplayTimeSorter(OrderSorter):
        def __init__(self, invert=True):
            self.invert = invert

        def create_prod_sequence(self, order):
            display_time = _get_order_time(order=order)

            if display_time and self.invert:
                future_date = iso8601.parse_date("3000-02-13T00:00:00Z")
                past_date = iso8601.parse_date("1984-02-13T00:00:00Z")
                last_change_date = iso8601.parse_date(display_time)

                sequence_date = (past_date + timedelta(seconds=(future_date - last_change_date).total_seconds()))
                return sequence_date.strftime("%Y%m%d%H%M%S")

            return display_time

    class DoneSorter(OrderSorter):
        def create_prod_sequence(self, order):
            return ""

    class InPreparationTimeSorter(OrderSorter):
        def __init__(
            self,
            invert=True,    # type: Optional[bool]
        ):
            # type: (...) -> None

            self.invert = invert

        def create_prod_sequence(
            self,
            order,  # type: ProductionOrder
        ):
            # type: (...) -> str

            in_preparation_time = _get_order_time(
                order=order,
                time_type="in_preparation",
            )

            if in_preparation_time and self.invert:
                future_date = iso8601.parse_date("3000-02-13T00:00:00Z")
                past_date = iso8601.parse_date("1984-02-13T00:00:00Z")
                last_change_date = iso8601.parse_date(in_preparation_time)

                sequence_date = (past_date + timedelta(seconds=(future_date - last_change_date).total_seconds()))
                return sequence_date.strftime("%Y%m%d%H%M%S")

            return in_preparation_time



from typing import Dict, Any
from datetime import datetime
from pos_model import Order
from dateutil import tz

from _SatXmlPartBuilder import SatXmlPartBuilder
from _ContextKeys import ContextKeys


class IdeBuilder(SatXmlPartBuilder):

    def __init__(self,
                 cnpj_sw, sign_ac):
        #  type: (unicode, unicode) -> None
        self.cnpj_sw = cnpj_sw
        self.sign_ac = sign_ac

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        last_totaled_state = None
        for order_state in reversed(order.states):
            if order_state.name == "TOTALED":
                last_totaled_state = order_state
                break

        if last_totaled_state is None:
            raise Exception("The order was not totaled")

        data_emissao = self._format_date(last_totaled_state.timestamp)

        context[ContextKeys.data_emissao] = data_emissao

        inner_xml = "<ide>"\
                    "<CNPJ>%s</CNPJ>"\
                    "<signAC>%s</signAC>"\
                    "<numeroCaixa>%03d</numeroCaixa>"\
                    "</ide>" % (self.cnpj_sw, self.sign_ac, int(order.pos_id))
        return inner_xml

    @staticmethod
    def _format_date(data):
        # type: (datetime) -> str
        local_zone = tz.tzlocal()
        data = data.replace(tzinfo=local_zone)
        data_str = data.strftime("%Y-%m-%dT%H:%M:%S%z")
        data_str = data_str[:22] + ":" + data_str[22:]
        return data_str



# cpf.py - functions for handling CPF numbers
# coding: utf-8
#
# Copyright (C) 2011-2016 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CPF (Cadastro de Pessoas Físicas, Brazilian national identifier).

The Cadastro de Pessoas Físicas is the Brazilian identification number
assigned to individuals for tax purposes. The number consists of 11 digits
and includes two check digits.

More information:

* https://en.wikipedia.org/wiki/Cadastro_de_Pessoas_Físicas

>>> validate('390.533.447-05')
'39053344705'
>>> validate('231.002.999-00')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('390.533.447=0')  # invalid delimiter
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> format('23100299900')
'231.002.999-00'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -.').strip()


def _calc_check_digits(number):
    """Calculate the check digits for the number."""
    d1 = sum((10 - i) * int(number[i]) for i in range(9))
    d1 = (11 - d1) % 11 % 10
    d2 = sum((11 - i) * int(number[i]) for i in range(9)) + 2 * d1
    d2 = (11 - d2) % 11 % 10
    return '%d%d' % (d1, d2)


def validate(number):
    """Check if the number is a valid CPF. This checks the length and whether
    the check digit is correct."""
    number = compact(number)
    if not isdigits(number) or int(number) <= 0:
        raise InvalidFormat()
    if len(number) != 11:
        raise InvalidLength()
    if _calc_check_digits(number) != number[-2:]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid CPF."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return number[:3] + '.' + number[3:6] + '.' + number[6:-2] + '-' + number[-2:]



# nif.py - functions for handling French tax identification numbers
# coding: utf-8
#
# Copyright (C) 2016-2021 Dimitri Papadopoulos
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NIF (Numéro d'Immatriculation Fiscale, French tax identification number).

The NIF (Numéro d'Immatriculation Fiscale or Numéro d'Identification Fiscale)
also known as numéro fiscal de référence or SPI (Simplification des
Procédures d'Identification) is a 13-digit number issued by the French tax
authorities to people for tax reporting purposes.

More information:

* https://ec.europa.eu/taxation_customs/tin/tinByCountry.html
* https://fr.wikipedia.org/wiki/Numéro_d%27Immatriculation_Fiscale#France

>>> validate('0701987765432')
'0701987765432'
>>> validate('070198776543')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('9701987765432')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> format('0701987765432')
'07 01 987 765 432'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip()


def validate(number):
    """Check if the number provided is a valid NIF."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if number[0] not in ('0', '1', '2', '3'):
        raise InvalidComponent()
    if len(number) != 13:
        raise InvalidLength()
    return number


def is_valid(number):
    """Check if the number provided is a valid NIF."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return ' '.join((number[:2], number[2:4], number[4:7],
                     number[7:10], number[10:]))



from ._ErrorType import ErrorType
from ._NfeTender import NfeTender
from ._PosTenderToNfeTender import PosTenderToNfeTender
from ._NfeIntermediary import NfeIntermediary
from ._NfePresence import NfePresence
from ._NfeCstatCode import NfceCstatCode
from ._FeeDetails import FeeDetails
from softwaredeveloper import SoftwareDeveloper



# postcode.py - functions for handling Dutch postal codes
#
# Copyright (C) 2013 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Postcode (the Dutch postal code).

The Dutch postal code consists of four numbers followed by two letters,
separated by a single space.

More information:
* https://en.wikipedia.org/wiki/Postal_codes_in_the_Netherlands
* https://nl.wikipedia.org/wiki/Postcodes_in_Nederland

>>> validate('2601 DC')
'2601 DC'
>>> validate('NL-2611ET')
'2611 ET'
>>> validate('26112 ET')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('2611 SS')  # a few letter combinations are banned
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""

import re

from stdnum.exceptions import *
from stdnum.util import clean


_postcode_re = re.compile(r'^(?P<pt1>[1-9][0-9]{3})(?P<pt2>[A-Z]{2})$')


_postcode_blacklist = ('SA', 'SD', 'SS')


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' -').upper().strip()
    if number.startswith('NL'):
        number = number[2:]
    return number


def validate(number):
    """Check if the number is in the correct format. This currently does not
    check whether the code corresponds to a real address."""
    number = compact(number)
    match = _postcode_re.search(number)
    if not match:
        raise InvalidFormat()
    if match.group('pt2') in _postcode_blacklist:
        raise InvalidComponent()
    return '%s %s' % (match.group('pt1'), match.group('pt2'))


def is_valid(number):
    """Check if the number is a valid postal code."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from msgbus import MBEasyContext
from persistence import (
    Driver,
    Connection,
    Cursor,
)

from typing import Optional


class ConnectionWrapper(object):
    def __init__(
        self,
        mb_context,         # type: MBEasyContext
        db_name=None,       # type: Optional[str]
        service_name=None,  # type: Optional[str]
    ):
        self.mb_context = mb_context
        self.db_name = db_name
        self.service_name = service_name
        self.conn = None  # type: Optional[Connection]

    def query(
        self,   # type: ConnectionWrapper
        sql,    # type: str
    ):
        # type: (...) -> int

        return self._get_conn().query(sql=sql)

    def select(
        self,   # type: ConnectionWrapper
        sql,    # type: str
    ):
        # type: (...) -> Cursor

        return self._get_conn().select(sql=sql)

    def begin_transaction(
        self,   # type: ConnectionWrapper
    ):
        # type: (...) -> None

        conn = self._get_conn()
        conn.transaction_start()
        conn.query(sql="BEGIN TRANSACTION")

    def commit(
        self,   # type: ConnectionWrapper
    ):
        # type: (...) -> None

        self.conn.query(sql="COMMIT")
        self.conn.transaction_end()

    def rollback(
        self,   # type: ConnectionWrapper
    ):
        # type: (...) -> None

        self.conn.query(sql="ROLLBACK")
        self.conn.transaction_end()

    def close(
        self,   # type: ConnectionWrapper
    ):
        # type: (...) -> None

        if self.conn is not None:
            self.conn.close()

    def _get_conn(
        self,   # type: ConnectionWrapper
    ):
        # type: (...) -> Connection

        if self.conn is None:
            self.conn = Driver().open(
                mbcontext=self.mb_context,
                dbname=str(self.db_name.encode("utf-8")) if self.db_name is not None else None,
                service_name=self.service_name.encode("utf-8"),
            )
        return self.conn



import json

from old_helper import BaseRepository
from typing import List, Optional

class OrderRepository(BaseRepository):
    def __init__(self, mb_context):
        super(OrderRepository, self).__init__(mb_context)

    def get_orders_benefit(self, initial_date, end_date, operator_id, pos_list):
        # type: (str, str, Optional[str], List[int]) -> List

        def inner_func(conn):

            query_operator = " and SessionId like '%user={0},%'".format(operator_id) if operator_id is not None else ""
            query = """SELECT o.OrderId, ocp.Value as BenefitList
                         FROM Orders o
                        INNER JOIN OrderTender ot
                           ON o.OrderId = ot.OrderId
                        INNER JOIN OrderCustomProperties ocp
                           ON ocp.OrderId = o.OrderId AND ocp.Key = 'BENEFIT_LIST'
                        WHERE o.BusinessPeriod >= '{0}' AND o.BusinessPeriod <= '{1}' AND o.StateId = 5
                        {2}
                        GROUP by o.OrderId
                        """.format(initial_date, end_date, query_operator)

            benefits = []
            cursor = conn.select(query)
            if cursor.rows() > 0:
                for row in cursor:
                    benefit_list_json = json.loads(row.get_entry("BenefitList"))
                    benefit_list_json = json.loads(benefit_list_json.get("benefitList", None))
                    for benefit in benefit_list_json:
                        benefit_applied = benefit.get("benefit_applied", False)
                        benefit_total_value = benefit.get("benefit_total_value", 0)
                        if not benefit_applied or benefit_total_value == 0:
                            continue
                        benefits.append(benefit)

            return benefits

        return self.execute_in_all_databases_returning_flat_list(inner_func, pos_list)

    def get_fiscalized_orders(
            self,
            from_date,  # type: str
            pos_list,   # type: List[int]
    ):
        # type: (...) -> List

        def inner_func(conn):
            query = """
               SELECT 
                    o.OrderId AS OrderId, 
                    ocp.Value AS FiscalizationDate, 
                    o.DistributionPoint AS PodType, 
                    O.OriginatorId AS OriginatorId, 
                    o.TotalGross AS TotalGross,
                    o.Tip AS Tip
               FROM Orders o
               JOIN OrderCustomProperties ocp ON 
                    o.OrderId = ocp.OrderId AND 
                    ocp.Key = 'FISCALIZATION_DATE'
                WHERE 
                    StateId = 5 AND
                    DATETIME(ocp.Value) >= DATETIME('{}')
            """.format(from_date)

            orders = []
            cursor = conn.select(query)
            if cursor.rows() > 0:
                for row in cursor:
                    fiscalized_order = FiscalizedOrder(
                        order_id=row.get_entry("OrderId"),
                        fiscalization_date=row.get_entry("FiscalizationDate"),
                        pod_type=row.get_entry("PodType"),
                        originator_id=row.get_entry("OriginatorId"),
                        total_gross=row.get_entry("TotalGross"),
                        tip=row.get_entry("Tip"),
                    )
                    orders.append(fiscalized_order)

            return orders

        return self.execute_in_all_databases_returning_flat_list(inner_func, pos_list)


class FiscalizedOrder(object):
    def __init__(
            self,
            order_id,               # type: str
            fiscalization_date,     # type: str
            pod_type,               # type: str
            originator_id,          # type: str
            total_gross,            # type: str
            tip,                    # type: str
    ):
        # type: (...) -> None
        self.order_id = order_id
        self.fiscalization_date = fiscalization_date
        self.pod_type = pod_type
        self.originator_id = originator_id
        self.total_gross = total_gross
        self.tip = tip



import json
from xml.etree import cElementTree as eTree

from product_api.dto import Part
from typing import (
    Dict,
    List,
)
from ._Order import Order
from ._SaleLine import SaleLine
from ._Tender import Tender
from ._CustomProperty import CustomProperty


def parse_order_custom_properties(
    order_xml,      # type: eTree.Element
):
    # type: (...) -> List[CustomProperty]

    custom_order_properties = order_xml.findall("CustomOrderProperties/OrderProperty")
    custom_properties = []    # type: List[CustomProperty]
    for custom_property in custom_order_properties:
        properties = custom_property.attrib
        if "key" not in properties:
            continue

        custom_properties.append(
            CustomProperty(
                key=properties.get("key"),
                value=properties.get("value"),
            )
        )

    return custom_properties


def parse_tenders(
    order_xml,      # type: eTree.Element
):
    # type: (...) -> List[Tender]

    tenders = order_xml.findall("TenderHistory/Tender")
    parsed_tenders = []     # type: List[Tender]
    for order_tender in tenders:
        parsed_tenders.append(
            Tender(
                tender_id=int(order_tender.get("tenderId", -1)),
                amount=float(order_tender.get("tenderAmount", 0))
            )
        )
    return parsed_tenders


def parse_sale_line_custom_properties(
    sale_line_xml,  # type: eTree.Element
):
    # type: (...) -> Dict[str, str]

    properties = sale_line_xml.get("customProperties")
    if not properties:
        return {}

    return json.loads(properties)


def parse_sale_line(
    sale_line_xml,  # type: eTree.Element
):
    # type: (...) -> SaleLine

    part = Part(
        part_code=int(sale_line_xml.get("partCode")),
        product_name=sale_line_xml.get("productName"),
        item_type=sale_line_xml.get("itemType"),
    )
    level = int(sale_line_xml.get("level") or 0)
    qty = float(sale_line_xml.get("qty") or 0) if level == 0 else float(sale_line_xml.get("multipliedQty") or 0)
    custom_properties = parse_sale_line_custom_properties(sale_line_xml=sale_line_xml)

    return SaleLine(
        line_number=int(sale_line_xml.get("lineNumber") or 0),
        part=part,
        level=level,
        item_id=sale_line_xml.get("itemId"),
        item_price=float(sale_line_xml.get("itemPrice") or 0),
        overwritten_unit_price=sale_line_xml.get("overwrittenUnitPrice"),
        quantity=qty,
        added_qty=float(sale_line_xml.get("addedQty") or 0),
        default_qty=int(sale_line_xml.get("defaultQty") or 0),
        comment=sale_line_xml.find("Comment"),
        custom_properties=custom_properties,
        discounts_applied=sale_line_xml.get("discountsApplied", None),
    )


def parse_order(
    order_xml,  # type: eTree.Element
):
    # type: (...) -> Order

    sale_lines_xml = order_xml.findall("SaleLine")
    sale_lines = []     # type: List[SaleLine]
    for sale_line in sale_lines_xml:
        sale_lines.append(parse_sale_line(sale_line_xml=sale_line))

    custom_properties = parse_order_custom_properties(order_xml=order_xml)

    tenders = parse_tenders(order_xml=order_xml)

    return Order(
        order_id=int(order_xml.get("orderId")),
        state=order_xml.get("state"),
        total_amount=float(order_xml.get("totalAmount") or 0),
        discount_amount=float(order_xml.get("discountAmount") or 0),
        total_after_discount=float(order_xml.get("totalAfterDiscount") or 0),
        total_tender=float(order_xml.get("totalTender") or 0),
        total_gross=float(order_xml.get("totalGross") or 0),
        sale_lines=sale_lines,
        custom_properties=custom_properties,
        tenders=tenders,
    )



from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder
from production.box import AllItemHasAtLeastOneTagFilterBox


class AllItemHasAtLeastOneTagFilterBoxBuilder(BoxBuilder):
    def __init__(self, loggers):
        super(AllItemHasAtLeastOneTagFilterBoxBuilder, self).__init__(loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> AllItemHasAtLeastOneTagFilterBox
        return AllItemHasAtLeastOneTagFilterBox(config.name,
                                                config.sons,
                                                self.get_extra(config, "Tags", []),
                                                self.get_logger(config))



# -*- coding: utf-8 -*-

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/type/quaternion.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1cgoogle/type/quaternion.proto\x12\x0bgoogle.type"8\n\nQuaternion\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\x12\t\n\x01z\x18\x03 \x01(\x01\x12\t\n\x01w\x18\x04 \x01(\x01\x42o\n\x0f\x63om.google.typeB\x0fQuaternionProtoP\x01Z@google.golang.org/genproto/googleapis/type/quaternion;quaternion\xf8\x01\x01\xa2\x02\x03GTPb\x06proto3'
)


_QUATERNION = DESCRIPTOR.message_types_by_name["Quaternion"]
Quaternion = _reflection.GeneratedProtocolMessageType(
    "Quaternion",
    (_message.Message,),
    {
        "DESCRIPTOR": _QUATERNION,
        "__module__": "google.type.quaternion_pb2"
        # @@protoc_insertion_point(class_scope:google.type.Quaternion)
    },
)
_sym_db.RegisterMessage(Quaternion)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\017com.google.typeB\017QuaternionProtoP\001Z@google.golang.org/genproto/googleapis/type/quaternion;quaternion\370\001\001\242\002\003GTP"
    _QUATERNION._serialized_start = 45
    _QUATERNION._serialized_end = 101
# @@protoc_insertion_point(module_scope)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from logging import getLogger
from re import split
from typing import Iterable, List, Mapping, Optional, Set
from urllib.parse import quote_plus, unquote_plus

from opentelemetry.baggage import _is_valid_pair, get_all, set_baggage
from opentelemetry.context import get_current
from opentelemetry.context.context import Context
from opentelemetry.propagators import textmap
from opentelemetry.util.re import _DELIMITER_PATTERN

_logger = getLogger(__name__)


class W3CBaggagePropagator(textmap.TextMapPropagator):
    """Extracts and injects Baggage which is used to annotate telemetry."""

    _MAX_HEADER_LENGTH = 8192
    _MAX_PAIR_LENGTH = 4096
    _MAX_PAIRS = 180
    _BAGGAGE_HEADER_NAME = "baggage"

    def extract(
        self,
        carrier: textmap.CarrierT,
        context: Optional[Context] = None,
        getter: textmap.Getter[textmap.CarrierT] = textmap.default_getter,
    ) -> Context:
        """Extract Baggage from the carrier.

        See
        `opentelemetry.propagators.textmap.TextMapPropagator.extract`
        """

        if context is None:
            context = get_current()

        header = _extract_first_element(
            getter.get(carrier, self._BAGGAGE_HEADER_NAME)
        )

        if not header:
            return context

        if len(header) > self._MAX_HEADER_LENGTH:
            _logger.warning(
                "Baggage header `%s` exceeded the maximum number of bytes per baggage-string",
                header,
            )
            return context

        baggage_entries: List[str] = split(_DELIMITER_PATTERN, header)
        total_baggage_entries = self._MAX_PAIRS

        if len(baggage_entries) > self._MAX_PAIRS:
            _logger.warning(
                "Baggage header `%s` exceeded the maximum number of list-members",
                header,
            )

        for entry in baggage_entries:
            if len(entry) > self._MAX_PAIR_LENGTH:
                _logger.warning(
                    "Baggage entry `%s` exceeded the maximum number of bytes per list-member",
                    entry,
                )
                continue
            if not entry:  # empty string
                continue
            try:
                name, value = entry.split("=", 1)
            except Exception:  # pylint: disable=broad-except
                _logger.warning(
                    "Baggage list-member `%s` doesn't match the format", entry
                )
                continue

            if not _is_valid_pair(name, value):
                _logger.warning("Invalid baggage entry: `%s`", entry)
                continue

            name = unquote_plus(name).strip()
            value = unquote_plus(value).strip()

            context = set_baggage(
                name,
                value,
                context=context,
            )
            total_baggage_entries -= 1
            if total_baggage_entries == 0:
                break

        return context

    def inject(
        self,
        carrier: textmap.CarrierT,
        context: Optional[Context] = None,
        setter: textmap.Setter[textmap.CarrierT] = textmap.default_setter,
    ) -> None:
        """Injects Baggage into the carrier.

        See
        `opentelemetry.propagators.textmap.TextMapPropagator.inject`
        """
        baggage_entries = get_all(context=context)
        if not baggage_entries:
            return

        baggage_string = _format_baggage(baggage_entries)
        setter.set(carrier, self._BAGGAGE_HEADER_NAME, baggage_string)

    @property
    def fields(self) -> Set[str]:
        """Returns a set with the fields set in `inject`."""
        return {self._BAGGAGE_HEADER_NAME}


def _format_baggage(baggage_entries: Mapping[str, object]) -> str:
    return ",".join(
        quote_plus(str(key)) + "=" + quote_plus(str(value))
        for key, value in baggage_entries.items()
    )


def _extract_first_element(
    items: Optional[Iterable[textmap.CarrierT]],
) -> Optional[textmap.CarrierT]:
    if items is None:
        return None
    return next(iter(items), None)



from typing import Dict, List, Protocol
from distutils.command.build import build as _build

from ..warnings import SetuptoolsDeprecationWarning

_ORIGINAL_SUBCOMMANDS = {"build_py", "build_clib", "build_ext", "build_scripts"}


class build(_build):
    # copy to avoid sharing the object with parent class
    sub_commands = _build.sub_commands[:]

    def get_sub_commands(self):
        subcommands = {cmd[0] for cmd in _build.sub_commands}
        if subcommands - _ORIGINAL_SUBCOMMANDS:
            SetuptoolsDeprecationWarning.emit(
                "Direct usage of `distutils` commands",
                """
                It seems that you are using `distutils.command.build` to add
                new subcommands. Using `distutils` directly is considered deprecated,
                please use `setuptools.command.build`.
                """,
                due_date=(2023, 12, 13),  # Warning introduced in 13 Jun 2022.
                see_url="https://peps.python.org/pep-0632/",
            )
            self.sub_commands = _build.sub_commands
        return super().get_sub_commands()


class SubCommand(Protocol):
    """In order to support editable installations (see :pep:`660`) all
    build subcommands **SHOULD** implement this protocol. They also **MUST** inherit
    from ``setuptools.Command``.

    When creating an :pep:`editable wheel <660>`, ``setuptools`` will try to evaluate
    custom ``build`` subcommands using the following procedure:

    1. ``setuptools`` will set the ``editable_mode`` attribute to ``True``
    2. ``setuptools`` will execute the ``run()`` command.

       .. important::
          Subcommands **SHOULD** take advantage of ``editable_mode=True`` to adequate
          its behaviour or perform optimisations.

          For example, if a subcommand doesn't need to generate an extra file and
          all it does is to copy a source file into the build directory,
          ``run()`` **SHOULD** simply "early return".

          Similarly, if the subcommand creates files that would be placed alongside
          Python files in the final distribution, during an editable install
          the command **SHOULD** generate these files "in place" (i.e. write them to
          the original source directory, instead of using the build directory).
          Note that ``get_output_mapping()`` should reflect that and include mappings
          for "in place" builds accordingly.

    3. ``setuptools`` use any knowledge it can derive from the return values of
       ``get_outputs()`` and ``get_output_mapping()`` to create an editable wheel.
       When relevant ``setuptools`` **MAY** attempt to use file links based on the value
       of ``get_output_mapping()``. Alternatively, ``setuptools`` **MAY** attempt to use
       :doc:`import hooks <python:reference/import>` to redirect any attempt to import
       to the directory with the original source code and other files built in place.

    Please note that custom sub-commands **SHOULD NOT** rely on ``run()`` being
    executed (or not) to provide correct return values for ``get_outputs()``,
    ``get_output_mapping()`` or ``get_source_files()``. The ``get_*`` methods should
    work independently of ``run()``.
    """

    editable_mode: bool = False
    """Boolean flag that will be set to ``True`` when setuptools is used for an
    editable installation (see :pep:`660`).
    Implementations **SHOULD** explicitly set the default value of this attribute to
    ``False``.
    When subcommands run, they can use this flag to perform optimizations or change
    their behaviour accordingly.
    """

    build_lib: str
    """String representing the directory where the build artifacts should be stored,
    e.g. ``build/lib``.
    For example, if a distribution wants to provide a Python module named ``pkg.mod``,
    then a corresponding file should be written to ``{build_lib}/package/module.py``.
    A way of thinking about this is that the files saved under ``build_lib``
    would be eventually copied to one of the directories in :obj:`site.PREFIXES`
    upon installation.

    A command that produces platform-independent files (e.g. compiling text templates
    into Python functions), **CAN** initialize ``build_lib`` by copying its value from
    the ``build_py`` command. On the other hand, a command that produces
    platform-specific files **CAN** initialize ``build_lib`` by copying its value from
    the ``build_ext`` command. In general this is done inside the ``finalize_options``
    method with the help of the ``set_undefined_options`` command::

        def finalize_options(self):
            self.set_undefined_options("build_py", ("build_lib", "build_lib"))
            ...
    """

    def initialize_options(self):
        """(Required by the original :class:`setuptools.Command` interface)"""

    def finalize_options(self):
        """(Required by the original :class:`setuptools.Command` interface)"""

    def run(self):
        """(Required by the original :class:`setuptools.Command` interface)"""

    def get_source_files(self) -> List[str]:
        """
        Return a list of all files that are used by the command to create the expected
        outputs.
        For example, if your build command transpiles Java files into Python, you should
        list here all the Java files.
        The primary purpose of this function is to help populating the ``sdist``
        with all the files necessary to build the distribution.
        All files should be strings relative to the project root directory.
        """

    def get_outputs(self) -> List[str]:
        """
        Return a list of files intended for distribution as they would have been
        produced by the build.
        These files should be strings in the form of
        ``"{build_lib}/destination/file/path"``.

        .. note::
           The return value of ``get_output()`` should include all files used as keys
           in ``get_output_mapping()`` plus files that are generated during the build
           and don't correspond to any source file already present in the project.
        """

    def get_output_mapping(self) -> Dict[str, str]:
        """
        Return a mapping between destination files as they would be produced by the
        build (dict keys) into the respective existing (source) files (dict values).
        Existing (source) files should be represented as strings relative to the project
        root directory.
        Destination files should be strings in the form of
        ``"{build_lib}/destination/file/path"``.
        """



import json

from interactor.message.get_operators_logout_sessions import GetOperatorsLogoutSessionsInteractor
from interactor.message.get_operators_logout_sessions.dto import GetOperatorsLogoutSessionsResponse
from messagebus import (
    DataType,
    DefaultToken,
    Event,
    Message,
    MessageBus,
)
from processor import BaseProcessor


class GetOperatorsLogoutSessionsProcessor(BaseProcessor):

    def __init__(
        self,           # type: GetOperatorsLogoutSessionsProcessor
        interactor,     # type: GetOperatorsLogoutSessionsInteractor
    ):
        # type: (...) -> None
        super(GetOperatorsLogoutSessionsProcessor, self).__init__()

        self.interactor = interactor

    def parse_data(
        self,           # type: GetOperatorsLogoutSessionsProcessor
        data,           # type: str
        event_type,     # type: str
    ):
        # type: (...) -> str

        return data

    def call_business(
        self,       # type: GetOperatorsLogoutSessionsProcessor
        request,    # type: str
    ):
        # type: (...) -> GetOperatorsLogoutSessionsResponse

        return self.interactor.execute(business_period=request)

    def format_response(
        self,           # type: GetOperatorsLogoutSessionsProcessor
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        request,        # type: str
        result,         # type: GetOperatorsLogoutSessionsResponse
    ):
        # type: (...) -> None

        response = {
            "business_period": result.business_period,
            "operators_logout_sessions": result.operators_logout_sessions,
        }

        reply = Message(
            token=DefaultToken.TK_SYS_ACK,
            data_type=DataType.string,
            data=json.dumps(response),
        )
        message_bus.reply_message(
            message=message,
            reply_message=reply,
        )



import os

import cfgtools

create_table_parallel_order = None


def can_create_table_parallel_order():
    # type: (...) -> str

    global create_table_parallel_order

    if create_table_parallel_order is None:
        config = cfgtools.read(os.environ["LOADERCFG"])
        create_table_parallel_order = config.find_value(
            "Customizations.CanCreateTableParallelOrder",
            "false"
        )

    return create_table_parallel_order.lower() == "true"



import re
from typing import (
    Dict,
    List,
    Optional,
    Union,
)

from nreports.repository import (
    I18NRepository,
    L10NRepository,
)
from nreports_api.shared import JSON
from ._Line import Line
from ._ReportFunction import ReportFunction
from .exceptions import ObjectListValueIsNotAList


class TextLine(Line):

    def __init__(
        self,
        line: JSON,
        text_attr: str,
    ) -> None:
        super().__init__(line)

        self._text_attr = text_attr
        self._variables: List[str] = []
        self._object_list_variables: List[str] = []
        self._functions: List[ReportFunction] = []
        self._i18n_terms: List[str] = []
        self._text_list: List[Line] = []

    def inner_format(
        self,
        data: Dict[str, JSON],
        i18n_repository: I18NRepository,
        l10n_repository: L10NRepository,
    ) -> List[Dict[str, Union[str, Dict[str, str]]]]:

        if "objectList" in self.line:
            formatted = self._format_object_list(object_lists=data)
        else:
            formatted = [
                {
                    self._text_attr: self.format_variables(
                        variables=data,
                        text=self.line[self._text_attr],
                    )
                }
            ]

            if "textList" in self.line:
                formatted[0]["textList"] = self._format_text_list(
                    data=data,
                    i18n_repository=i18n_repository,
                    l10n_repository=l10n_repository,
                )

        for line in formatted:
            text = self._format_i18n_labels(
                text=line[self._text_attr],
                i18n_repository=i18n_repository,
            )
            line[self._text_attr] = self._format_functions(
                text=text,
                l10n_repository=l10n_repository,
            )

            if not line[self._text_attr]:
                line["command"] = "{CMD_BREAK_LINE}"
                line.pop(self._text_attr)
            elif "customization" in self.line:
                line["customization"] = self.line.get("customization")

        return formatted

    def _format_text_list(
        self,
        data: Dict[str, JSON],
        i18n_repository: I18NRepository,
        l10n_repository: L10NRepository,
    ) -> List[Dict[str, str]]:

        formatted_text_list: List[Dict[str, str]] = []
        for text_line in self._text_list:
            formatted_text_list.extend(
                text_line.format(
                    data=data,
                    i18n_repository=i18n_repository,
                    l10n_repository=l10n_repository,
                )
            )

        return formatted_text_list

    def set_text_list(
        self,
        text_list: List[Line],
    ) -> None:

        self._text_list = text_list

    def _format_object_list(
        self,
        object_lists: Dict[str, List[Dict[str, JSON]]],
    ) -> List[Dict[str, str]]:

        object_list_values = object_lists[self.line["objectList"]]
        if not isinstance(object_list_values, list):
            raise ObjectListValueIsNotAList(name=self.line["objectList"])

        formatted_lines = []
        for line_variables in object_list_values:
            formatted_line = self.format_variables(
                variables=line_variables,
                text=self.line[self._text_attr],
            )
            split_lines = formatted_line.split("\n")
            for line in split_lines:
                formatted_lines.append({self._text_attr: line})

        return formatted_lines

    def get_inner_data_to_build(
        self,
    ) -> Dict[str, Optional[Dict[str, None]]]:

        ret: Dict[str, Optional[Dict[str, None]]] = {}
        for variable in self.get_text_variables(text=self.line[self._text_attr]):
            ret[variable[1:-1]] = None

        if "objectList" in self.line:
            inner = ret
            ret = {
                self.line["objectList"]: inner,
            }

        if "textList" in self.line:
            for text_line in self._text_list:
                ret.update(text_line.get_inner_data_to_build())

        return ret

    @staticmethod
    def _format_functions(
        text: str,
        l10n_repository: L10NRepository,
    ) -> str:

        functions = TextLine._get_function_sentences(text=text)

        for function_sentence in functions:
            report_function = ReportFunction(
                function_sentence=function_sentence,
                l10n_repository=l10n_repository,
            )
            text = text.replace(
                function_sentence,
                report_function.execute(parsed_params={}),
            )

        return text

    @staticmethod
    def _get_function_sentences(
        text: str,
    ) -> List[str]:

        functions = []
        inner_funcs = 0
        inside_function = False
        function_sentence = ""
        for i in range(len(text)):
            c = text[i]
            if c == "#":
                inside_function = True

            if not inside_function:
                continue

            if c == "(":
                inner_funcs += 1
            elif c == ")":
                inner_funcs -= 1

                if inner_funcs == 0:
                    function_sentence += c
                    functions.append(function_sentence)
                    function_sentence = ""
                    inside_function = False

            if inside_function:
                function_sentence += c

        if function_sentence != "":
            functions.append(function_sentence)

        return functions

    @staticmethod
    def _format_i18n_labels(
        text: str,
        i18n_repository: I18NRepository,
    ) -> str:

        all_keys = re.findall(r"\$\w+", text)
        translations = i18n_repository.translate_labels(labels=all_keys)
        for i18n_key in all_keys:
            if i18n_key in translations:
                text = text.replace(
                    i18n_key,
                    translations[i18n_key],
                )

        return text



from typing import (
    Dict,
    Any,
    List,
    Union,
    Optional,
)

from nreports.builders import DataBuilder
from nreports.model import GroupedSaleItems
from nreports.services import PreAccountOrdersRetrieverService
from nreports.utils import convert_from_utf_to_localtime
from nreports_api.shared import JSON
from order_api.dto import (
    SaleLine as SaleLineApi,
    Order as OrderApi,
)
from table_manager_api.dto import (
    Table,
    ServiceTender,
)


class PreAccountBuilder(DataBuilder):

    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        pos_id = int(kwargs["posId"])
        table_id = kwargs.get("tableId")

        self._build_mobile_variables(
            data_to_build=data_to_build,
            **kwargs,
        )

        order_retriever_service = PreAccountOrdersRetrieverService(
            message_bus=self._message_bus,
            user_repository=self.user_repository,
            store_config_repository=self.store_config_repository,
        )
        table = order_retriever_service.get_table_order(
            pos_id=pos_id,
            table_id=table_id,
        )

        self._build_header(
            data_to_build=data_to_build,
            table=table,
        )

        self._build_body(
            pos_id=pos_id,
            data_to_build=data_to_build,
            table=table,
        )

        self._build_totals(
            data_to_build=data_to_build,
            table=table,
        )

        self._build_footer(data_to_build=data_to_build)

        return data_to_build

    def _build_mobile_variables(
        self,
        data_to_build: JSON,
        **kwargs: Any,
    ) -> None:

        is_mobile = str(kwargs.get("isMobile", "false")).lower() == "true"
        data_to_build["IS_MOBILE"] = is_mobile
        data_to_build["NOT_IS_MOBILE"] = not is_mobile
        if is_mobile:
            is_small = self.constants.get("MOBILE_PRE_ACCOUNT_FONT_SIZE", "small").lower() == "small"
            data_to_build["IS_SMALL_FONT"] = is_small
            data_to_build["IS_LARGE_FONT"] = not is_small

    @staticmethod
    def _build_header(
        data_to_build: JSON,
        table: Table,
    ) -> None:

        show_service_seats = True if table.table_id else False
        data_to_build["TABLE_NAME"] = table.table_name
        data_to_build["ORDER_ID"] = table.order_id
        data_to_build["SHOW_ORDER_ID"] = True
        data_to_build["SHOW_SERVICE_SEATS"] = show_service_seats
        data_to_build["USER_NAME"] = table.user_name
        data_to_build["SERVICE_SEATS"] = table.service_seats
        data_to_build["START_DATE"] = convert_from_utf_to_localtime(utc=table.start_date)

    def _build_body(
        self,
        pos_id: int,
        data_to_build: JSON,
        table: Table,
    ) -> None:

        grouped_sale_lines = self._get_simplified_sale_lines(table=table)
        products_lines = self._build_product_lines(sales_line_list=grouped_sale_lines)
        data_to_build["ITEM_PRODUCT"] = products_lines

        currency_symbol = self.i18n_repository.build_l10n_repository(pos_id=pos_id).get_currency_symbol()
        data_to_build["CURRENCY_SYMBOL"] = currency_symbol.strip()

    def _get_simplified_sale_lines(
        self,
        table: Table,
    ) -> List[SaleLineApi]:

        grouped_sale_lines = self._get_grouped_sale_lines(table_orders=table.orders)
        fraction_line_numbers = []
        for _, sale_lines in grouped_sale_lines.items():
            fraction_line_numbers.extend(self._get_fraction_sale_line_identifiers(sales_line_list=sale_lines))

        all_orders_sale_lines = []
        for order in table.orders:
            all_orders_sale_lines.extend(order.sale_lines)

        return GroupedSaleItems().grouped_sale_lines(
            sale_lines=all_orders_sale_lines,
            hide_items_without_price=True,
            fraction_line_numbers=fraction_line_numbers,
        )

    def _build_totals(
        self,
        data_to_build: JSON,
        table: Table,
    ) -> None:

        self._build_amount_fields(
            table=table,
            data_to_build=data_to_build,
        )

        self._build_discount_fields(
            discount_groups=table.discount_groups,
            total_discounts=table.total_discounts,
            data_to_build=data_to_build,
        )

        self._build_payments_fields(
            table=table,
            data_to_build=data_to_build,
        )

        self._build_tip_selection(
            table=table,
            data_to_build=data_to_build,
        )

        self._build_division_suggestion(
            table=table,
            data_to_build=data_to_build,
        )

    def _build_footer(
        self,
        data_to_build: JSON,
    ) -> None:

        final_message = self.constants.get("PRE_ACCOUNT_FINAL_MESSAGE")

        if final_message == "":
            final_message = u""
        elif final_message is None:
            final_message = "$SERVICE_TAX_IS_NOT_MANDATORY"

        data_to_build["PRE_ACCOUNT_FINAL_MESSAGE"] = final_message

    def _build_discount_fields(
        self,
        discount_groups: Dict[str, float],
        total_discounts: float,
        data_to_build: JSON,
        list_name: Optional[str] = "INNER_LIST_DISCOUNT",
        condition_name: Optional[str] = "HAS_DISCOUNT_LIST",
    ) -> None:

        data_to_build[condition_name] = False
        if total_discounts <= 0:
            return

        data_to_build[condition_name] = True
        total_discount_groups: Dict[str, float] = {}
        for group in discount_groups:
            if group not in total_discount_groups:
                total_discount_groups[group] = 0

            total_discount_groups[group] += discount_groups[group]

        has_groups = len(total_discount_groups) > 0
        if has_groups:
            list_discount = self._parse_discounts_group(discount_groups=total_discount_groups)
        else:
            list_discount = [{
                "DISCOUNT_LABEL": "$DISCOUNTS",
                "DISCOUNT_VALUE": -total_discounts,
            }]

        data_to_build[list_name] = list_discount

    def _build_payments_fields(
        self,
        table: Table,
        data_to_build: JSON,
    ) -> None:

        data_to_build["HAS_SERVICE_TENDERS"] = False
        if not table.service_tenders:
            return

        data_to_build["HAS_SERVICE_TENDERS"] = True
        service_tenders = table.service_tenders
        data_to_build["INNER_LIST_PARTIAL_PAYMENTS"] = self._parse_partial_payments(service_tenders=service_tenders)
        total_paid = sum([float(service_tender.amount) for service_tender in service_tenders])
        data_to_build["TO_PAY"] = table.total_amount - total_paid

    def _build_tip_selection(
        self,
        table: Table,
        data_to_build: JSON,
    ) -> None:

        data_to_build["HAS_SERVICE_SEATS"] = False
        if not table.service_seats:
            return

        available_tips = self.store_config_repository.get_available_tips()
        if not available_tips:
            return

        data_to_build["HAS_SERVICE_SEATS"] = True
        sub_total_amount = table.sub_total_amount
        tip_fields: List[Dict[str, str]] = []
        for tip in available_tips:
            tip_amount = sub_total_amount * (float(tip) / 100)
            total_amount = sub_total_amount + tip_amount
            tip_fields.append(
                {
                    "TIP": tip,
                    "TIP_AMOUNT": tip_amount,
                    "TOTAL_AMOUNT": total_amount
                }
            )

        data_to_build["ITEM_TIP_SELECTION"] = tip_fields

    @staticmethod
    def _build_amount_fields(
        table: Table,
        data_to_build: JSON,
    ) -> None:

        data_to_build["ORDER_TOTAL_GROSS"] = table.sub_total_amount

        data_to_build["HAS_PROMOTIONAL_TENDERS"] = table.promotional_tenders > 0
        data_to_build["PROMOTIONAL_TENDERS"] = table.promotional_tenders

        data_to_build["HAS_TIP_AMOUNT"] = table.tip > 0
        data_to_build["TOTAL_WITHOUT_TIP"] = table.sub_total_amount - table.total_discounts
        data_to_build["TIP_AMOUNT"] = table.tip
        data_to_build["TOTAL_AMOUNT"] = table.total_amount

    @staticmethod
    def _build_division_suggestion(
        table: Table,
        data_to_build: JSON,
    ) -> None:

        data_to_build["HAS_DIVISION_SUGGESTION"] = False
        if table.service_seats < 1:
            return

        data_to_build["HAS_DIVISION_SUGGESTION"] = True
        division_amount = table.total_amount / table.service_seats
        data_to_build["SERVICE_SEATS"] = table.service_seats
        data_to_build["DIVISION_AMOUNT"] = division_amount

    @staticmethod
    def _build_product_lines(
        sales_line_list: List[SaleLineApi],
    ) -> List[Dict[str, str]]:

        item_products: List[Dict[str, str]] = []
        for item in sales_line_list:
            item_service_period = f"[{item.service_period}]" if item.service_period else " "

            item_price = item.item_price or 0
            item_price = f"#NUMBER({item_price})" if item_price else ""

            comment_description = ""
            if item.comment and item.comment.startswith("["):
                comment_description = "$" + item.comment.replace("[", "").replace("]", "").upper() + " "

            product_description = comment_description + item.part.product_name
            fraction = item.fraction_parent
            if fraction:
                numerator = int(round(item.quantity * fraction))
                qty_description = f" {numerator}/{fraction}" if numerator != fraction else "1.00"
            else:
                qty_description = "{:.2f}".format(item.quantity)

            indentation = " " * (item.level + 1)
            description = f"{indentation}{product_description}"
            item_products.append(
                {
                    "ITEM_PRODUCT_QUANTITY": qty_description,
                    "ITEM_PRODUCT_NAME": description,
                    "ITEM_SERVICE_PERIOD": item_service_period,
                    "ITEM_PRODUCT_TOTAL_PRICE": item_price,
                }
            )

        return item_products

    @staticmethod
    def _get_fraction_sale_line_identifiers(
        sales_line_list: List[SaleLineApi],
    ) -> List[str]:

        fraction_sale_lines_list = []
        for sale_line in sales_line_list:
            if sale_line.fraction_parent:
                order_id = sale_line.get_custom_property(key="orderId")
                line_number_identifier = f"{order_id}-{sale_line.line_number}"
                fraction_sale_lines_list.append(line_number_identifier)

        return fraction_sale_lines_list

    def _get_grouped_sale_lines(
        self,
        table_orders: List[OrderApi],
    ) -> Dict[Union[Union[int, str], Any], List[SaleLineApi]]:

        grouped_sale_lines = {}
        for order in table_orders:
            for sale_line in order.sale_lines:
                if self._is_invalid_sale_line(sale_line=sale_line):
                    continue

                seat = self.get_table_seat(
                   sale_lines=order.sale_lines,
                   sale_line=sale_line,
                )
                if not seat:
                    continue

                if seat not in grouped_sale_lines:
                    grouped_sale_lines[seat] = [sale_line]
                    continue

                grouped_sale_lines[seat].append(sale_line)

        return grouped_sale_lines

    def get_table_seat(
        self,
        sale_lines: List[SaleLineApi],
        sale_line: SaleLineApi,
    ) -> str:

        seat = sale_line.seat
        if sale_line.level != 0:
            seat = self._get_parent_table_seat(
                sale_lines=sale_lines,
                line_number=sale_line.line_number,
            )

        return "$TABLE" if seat == 0 else seat

    @staticmethod
    def _get_parent_table_seat(
        sale_lines: List[SaleLineApi],
        line_number: int,
    ) -> int:
        seat = [x.seat for x in sale_lines if x.level == 0 and x.line_number == line_number]
        return seat[0] if len(seat) > 0 else 0

    @staticmethod
    def _parse_discounts_group(
        discount_groups: Dict[str, float],
    ) -> List[Dict[str, str]]:

        total_discounts = 0
        discount_field: List[Dict[str, str]] = []
        for discount in discount_groups:
            value = float(discount_groups.get(discount))
            if value == 0:
                continue

            discount_field.append({
                "DISCOUNT_LABEL": discount,
                "DISCOUNT_VALUE": -value,
            })
            total_discounts += value

        discount_field.append({
            "DISCOUNT_LABEL": "$TOTAL",
            "DISCOUNT_VALUE": -total_discounts,
        })

        return discount_field

    @staticmethod
    def _parse_partial_payments(
        service_tenders: List[ServiceTender],
    ) -> List[Dict[str, str]]:

        payments_field: List[Dict[str, str]] = []
        for tender in service_tenders:
            payments_field.append({
                "PAYMENT_LABEL": tender.description,
                "PAYMENT_VALUE": tender.amount,
            })

        total_paid = sum([float(tender.amount) for tender in service_tenders])
        payments_field.append({
            "PAYMENT_LABEL": "$PAID",
            "PAYMENT_VALUE": total_paid,
        })

        return payments_field

    @staticmethod
    def _is_invalid_sale_line(
        sale_line: SaleLineApi,
    ) -> bool:

        item_type = sale_line.part.item_type
        return item_type == "COUPON" or (item_type == "COMBO" and sale_line.quantity == 0)
